.TH FITS.pm 3 2004.07.01
.SH NAME
FITS.pm \- a perl module for working with FITS files
.SH SYNOPSIS
This FITS module can be included in any perl program by including the following
line.

.B "    require FITS;"

There are two ways to call functions included in the module, depending on
whether or not the function requires a FITS header object. To call a function
which doesn't require a FITS header object, use the following style:

.B "    FITS::function_name ([$param1, [...]]);"

To call a function which does require a FITS header object, use the following
style, (assuming $fitsObj is a reference to the FITS object.)

.B "    $fitsObj->function_name ([$param1, [...]]);"
.SH DESCRIPTION
The following sections describe the various types of functions provided by this
module. The sections are split up into functions for creating FITS header
objects, getting information from a FITS header object, modifying FITS header 
objects, and writing FITS header objects to files.
.P
There are two main types of objects in the FITS package, headers and cards.
Headers contain a list of cards, and can be associated with a data array which
will be appended to the header when writing to a file. Cards are fairly simple,
and contain a keyword and value along with an optional comment. Commentary
cards (whose keywords are either "COMMENT", "HISTORY" or "") cannot have
comments associated with them. Functions which require a FITS object will
require either a header or card object. The type required will be mentioned
below in the description for each function.
.SH CREATING FITS HEADER OBJECTS
FITS header objects are the highest-level building blocks for dealing with
FITS files. To append a data array to a FITS header object, see the section
.B DATA ARRAYS
later in this documentation.
.SS
.BI "FITS::create (" $fitsFilename ")"
The
.B create
function, which does not require a FITS header object, allows you to create a
FITS header object and possibly fill it with data read from the file named
.IR $fitsFilename ". If " $fitsFilename
isn't provided, then the FITS header object will be filled with the default
values for primary headers.
.TP
.B Parameters:
.IR $fitsFilename " - (optional) - the filename of the fits file which"
contains data to read into the FITS header object.
.TP
.B Returns:
A FITS header object, filled with data from the FITS file
.I $fitsFilename
if provided.
.SS
.BI "FITS::build_header (" $extendedFlag ")"
The
.B build_header
function, which does not require a FITS header object, allows you to create a
default FITS header with the appropriate keywords for primary or extended
header.
.TP
.B Parameters:
.IR $extendedFlag " - (optional) - This parameter, if it contains a value"
which evaluates to true, tells
.B build_header
to construct an extended header instead of a primary header. The header will
be filled with the default keywords for primary or extended headers, depending
on the value of
.IR $extendedFlag .
.TP
.B Returns:
A primary or extended FITS header object with required keywords filled with
default values.
.P
** These two functions allow for the creation of blank primary and extended
headers, as well as reading the contents for a primary header from a file. The
reason there is no way to read an extended header from a file is because all
FITS files must begin with a primary header. Support for extended headers isn't
fully supported yet, but in later versions the extended headers will be read
when the primary header is read from a file, and will be accessable from the
FITS primary header object.
.SH READING FITS HEADER OBJECTS
.SS
.IB $fitsObj "->get_card (" $keyword " [, " $index "])"
This function returns a FITS card object associated with the keyword
.IR $keyword .
If
.I $index
is given and multiple cards exist with the given keyword,
.I $index
determines which will be returned (1 for the first, 2 for the second, etc.)
.I $fitsObj
must be a header object.
.SS
.IB $fitsObj "->get_value ()"
This function returns the value associated with a FITS card object.
.I $fitsObj
must be a card.
.SS
.IB $fitsObj "->{KEYWORD}"
This is not a function, but it returns the keyword associated with a fits
card object.
.I $fitsObj
must be a FITS card object.
.SS
.IB $fitsObj "->{COMMENT}"
This is also not a function, but it returns the comment associated with a fits
card object.
.I $fitsObj
must be a FITS card object.
.SH MODIFYING FITS HEADER OBJECTS
Currently, modifying a FITS header object entails only adding cards. A card is
usually a keyword/value pair with an optional comment, but commentary cards
(in which the keyword is "COMMENT", "HISTORY" or blank) cannot have a comment.
.SS
.BI "$fitsObj->set_card (" $keyword ", " $value " [, " $comment "])"
The
.B set_card
function, which requires a fits header object ($fitsobj in the above example)
adds a card with the given keyword, value, and optional comment to the FITS
header. In the case that the keyword is a required keyword for the header being
modified, rather than adding a new keyword with the same name, which isn't
allowed by the FITS standard, the card with that keyword is modified.

In the case that the keyword is NAXIS, which is required by both primary and
extended headers and determines the number of NAXISn keywords present, then
the appropriate number of NAXISn cards will be added or removed. For example,
if NAXIS is set to 2, then two cards named NAXIS1 and NAXIS2 will be added
right after the NAXIS keyword, and will be set with the default values of 0.
If later, NAXIS is set to 1, the NAXIS2 keyword will be removed so that there
is only one NAXISn keyword.
.SH WRITING FITS HEADERS TO FILES
.SS
.BI "$fitsObj->write_to_file (" $filename ")"
The
.B write_to_file
function, which requires a fits header object, writes the contents of the
FITS header object to a file named
.IR $filename .
Any data array associated with the header will also be added.
.SH DATA ARRAYS
.SS
.BI "$fitsObj->set_data (" $filename " [, " $offset " [, " $recordBytes "]])"
The
.B set_data
method allows you to specify the data array to append to the header. The size
of the data array is computed directly from the BITPIX, NAXIS, and NAXISn
keywords, so all you need to specify is the beginning of the data array. You do
this by providing a file, followed by the location within the file. If the
location is ommitted then it is assumed that the data array is located at the 
beginning of the file. If
.I $offset
is given, and
.I $recordBytes
is not, then the data is assumed to begin at byte
.I $offset
within the file. If
.IR $offset " and " $recordBytes
are both given, then the data is assumed to begin at record
.I $offset
where the record size (in bytes) is given by
.IR $recordBytes .
.I $offset
starts at one, not zero, for the first byte or record
.B Examples:
.IP
.B $fitsObj->set_data ("file.fits")
.br
will append the contents of the file named "file.fits".
.IP
.B $fitsObj->set_data ("file.fits", 1024)
.br
will append the contents of the file named "file.fits", starting at the 1024th
byte.
.IP
.B $fitsObj->set_data ("file.fits", 3, 512)
.br
will append the contents of the file name "file.fits", starting at the 3rd
block of 512 bytes (or 1636 bytes in.)