#!/usr/bin/perl

# declare package name
package FITS;

# declare some constants
use constant FALSE => 0;
use constant TRUE  => not FALSE;
use constant CARDIMAGELENGTH =>   80;
use constant RECORDSIZE      => 2880;
use constant CARDSPERRECORD  =>   36; # 2880 / 80

# define list of object types
my $types = {ERROR            =>  0,
	     PRIMARY_HEADER   =>  1,
	     EXTENDED_HEADER  =>  2,
	     # Card/Keyword Types
	     NAMED_CARD       =>  3,
	     END_CARD         =>  4,
	     COMMENT_CARD     =>  5,
	     HISTORY_CARD     =>  6,
	     BLANK_CARD       =>  7,
	     # Value Types
	     UNKNOWN_TYPE     =>  8,
	     CHARACTER_STRING =>  9,
	     LOGICAL_VARIABLE => 10,
	     INTEGER          => 11,
	     COMPLEX_INTEGER  => 12,
	     REAL             => 13,
	     COMPLEX_REAL     => 14};

# The create(<$filename>) function creates a FITS primary header object, see
# the FITS man page for notes on usage.
sub create (;$)
{
    # get parameters
    my ($filename) = @_;
    # create self
    my $self = build_header ();
    # load from file if filename given
    $self = $self->load_from_file ($filename) if ($filename);
    # return self
    return $self;
}

# The build_header function creates a FITS object, which is either a
# primary header if there are no pararmeters given, or the first parameter
# given resolves to false, and is an extended header if the first parameter
# given resolves to true. See the FITS man page for documentation on usage.
sub build_header ($)
{
    # get parameters
    my ($extendedHeaderFlag) = @_;
    # create header object
    my $header = bless ({});
    # set type depending on the $extendedHeaderFlag parameter
    $header->{TYPE} = ($extendedHeaderFlag?
		       $types->{EXTENDED_HEADER}:
		       $types->{PRIMARY_HEADER});
    # fill with default values for this type of header
    $header->fill_default_values ();
    # return header object
    return $header;
}

# The fill_default_values function fills the header (primary or extended) with
# the default values for that header type. It just calls 
# fill_primary_header_values of fill_extended_header_values depending on the
# header type. See the FITS man page for documentation on usage.
sub fill_default_values ()
{
    # get self from parameter list
    my $self = shift;
    # determine type of header, and fill appropriately
    if ($self->{TYPE} == $types->{PRIMARY_HEADER})
    {
	# type is PRIMARY_HEADER, fill with primary header values
	return $self->fill_primary_header_values ();
    }
    elsif ($self->{TYPE} == $types->{EXTENDED_HEADER})
    {
	# type is EXTENDED_HEADER, fill with extended header values
	return $self->fill_extended_header_values ();
    }
    # typs is not a header, return failure
    return undef;
}

# The fill_primary_header_values function fills a header with the default
# keywords and values for primary headers. See the FITS man page for 
# documentation on usage.
sub fill_primary_header_values ()
{
    # get self from parameter list
    my $self = shift;
    # add default cards and values for a primary header
    push (@{$self->{CARDS}}, build_named_card ("SIMPLE", "T"));
    push (@{$self->{CARDS}}, build_named_card ("BITPIX", 8));
    push (@{$self->{CARDS}}, build_named_card ("NAXIS",  0));
    # success
    return TRUE;
}

# The fill_extension_header_values function fills a header with the default
# keywords and values for extended headers. See the FITS man page for
# documentation on usage.
sub fill_extended_header_values ()
{
    # get self from parameter list
    my $self = shift;
    # add default cards and values for an extended header
    push (@{$self->{CARDS}}, build_named_card ("XTENSION", "''"));
    push (@{$self->{CARDS}}, build_named_card ("BITPIX", 8));
    push (@{$self->{CARDS}}, build_named_card ("NAXIS",  0));
    push (@{$self->{CARDS}}, build_named_card ("PCOUNT", 0));
    push (@{$self->{CARDS}}, build_named_card ("GCOUNT", 0));
    # success
    return TRUE;
}

# The set_card function creates (or updates for reserved keywords) cards within
# a FITS header object. See the FITS man page for documentation on usage.
sub set_card ($$;$)
{
    # get parameters
    my ($self, $keyword, $value, $comment) = @_;
    # declare local variables
    my ($card, @cards);
    # make sure object is header
    return undef unless ($self->is_header ());
    # check keyword for special values
    if ($keyword =~ /^comment\s*$/i)
    {
	# build a COMMENT card
	$card = build_comment_card ($value);
    }
    elsif ($keyword =~ /^history\s*$/i)
    {
	# build a HISTORY card
	$card = build_history_card ($value);
    }
    elsif ($keyword =~ /^end\s*$/i)
    {
	# build an END card
	$card = build_end_card ();
    }
    elsif ($keyword =~ /^\s*$/i)
    {
	# build a card with no keyword (1st 8 chars are spaces.)
	$card = build_blank_card ($value);
    }
    # These next few cases handle situations where the cards being set are
    # reserved keywords for the header type (primary or extended.) Rather than
    # add new cards, the existing cards are modified, and any other required
    # changes to the FITS structure are made.
    elsif ($self->{TYPE} == $types->{PRIMARY_HEADER} and
	   $keyword eq "SIMPLE")
    {
	# Get first card, must be SIMPLE for primary header
	$card = $self->{CARDS}->[0];
	# make sure the 1st card was, in fact, a SIMPLE card
	die ("First card of primary header not SIMPLE?")
	    unless ($card->{KEYWORD} eq "SIMPLE");
	# build and set the new value
	$card->{VALUE} = build_value_object ($value);
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $card;
    }
    elsif ($self->{TYPE} == $types->{EXTENDED_HEADER} and
	   $keyword eq "XTENSION")
    {
	# Get first card, must be XTENSION for extended header
	$card = $self->{CARDS}->[0];
	# make sure the 1st card was, in fact, an XTENSION card
	die ("First card of extension header not XTENSION?")
	    unless ($card->{KEYWORD} eq "XTENSION");
	# build and set the new value
	$card->{VALUE} = build_value_object ($value);
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $card;
    }
    elsif ($self->{TYPE} == $types->{EXTENDED_HEADER} and
	   $keyword eq "PCOUNT")
    {
	# Retrieve the PCOUNT card
	@cards = grep {$_->{KEYWORD} eq "PCOUNT";} @{$self->{CARDS}};
	# make sure it was there
	die ("PCOUNT not present in extended header") if ($#cards != 0);
	# build and set the new value
	$cards[0]->{VALUE} = build_value_object ($value);
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $cards[0];
    }
    elsif ($self->{TYPE} == $types->{EXTENDED_HEADER} and
	   $keyword eq "GCOUNT")
    {
	# Retrieve the GCOUNT card
	@cards = grep {$_->{KEYWORD} eq "GCOUNT";} @{$self->{CARDS}};
	# make sure it was there
	die ("GCOUNT not present in extended header") if ($#cards != 0);
	# build and set the new value
	$cards[0]->{VALUE} = build_value_object ($value);
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $cards[0];
    }
    elsif ($keyword eq "BITPIX")
    {
	# get the BITPIX card, must be the second card
	$card = $self->{CARDS}->[1];
	# make sure it was there
	die ("Second card not BITPIX")
	    unless ($card->{KEYWORD} eq "BITPIX");
	# build and set the new value
	$card->{VALUE} = build_value_object ($value);
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $card;
    }
    elsif ($keyword eq "NAXIS")
    {
	# get the NAXIS card, must be the third card
	$card = $self->{CARDS}->[2];
	# make sure it was there
	die ("Third card not NAXIS")
	    unless ($card->{KEYWORD} eq "NAXIS");
	# build and set the new value
	$card->{VALUE} = build_value_object ($value);
	# Add any new NAXISn cards:
	# For every number n between 1 and the new NAXIS value, inclusive...
	for ($index = 1; $index <= $value; ++$index)
	{
	    # if an NAXISn card doesn't exist
	    if (2 + $index > $#{@{$self->{CARDS}}} or
		$self->{CARDS}->[2 + $index]->{KEYWORD} ne "NAXIS".$index)
	    {
		# build the NAXISn card, default its value to zero
		$naxisN = build_named_card ("NAXIS".$index, 0);
		# insert the card in the appropriate location in the header
		splice (@{$self->{CARDS}}, 2 + $index, 0, $naxisN);
	    }
	}
	# Remove any extra NAXISn cards:
	# while there are more NAXISn cards (where n is greater than NAXIS)
	while (2 + $index <= $#{@{$self->{CARDS}}} and
	       $self->{CARDS}->[2 + $index]->{KEYWORD} =~ /^NAXIS\d{1,3}$/)
	{
	    # remove the NAXISn card
	    splice (@{$self->{CARDS}}, 2 + $index, 1);
	}
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $card;
    }
    elsif ($keyword =~ /^NAXIS(\d{1,3})$/)
    {
	# get the 'n' number for NAXISn from the regex above
	$index = $1;
	# Get the appropriate card
	$card = $self->{CARDS}->[2 + $index];
	# make sure it was where we expected it to be
	die ("NAXIS$index not found where expected\n".
	     "Did you set NAXIS first?")
	    unless ($card->{KEYWORD} eq "NAXIS".$index);
	# build and set the new value
	$card->{VALUE} = build_value_object ($value);
	# set the comment if given
	$card->{COMMENT} = $comment if $comment;
	# finished
	return $card;
    }
    # Finished with reserved keywords
    elsif (validate_keyword ($keyword))
    {
	# $keyword is valid, build a new card
	$card = build_named_card ($keyword, $value, $comment);
    }
    else
    {
	# $keyword is invalid, print warning and return failure
	warn ("Warning: illegal FITS keyword $keyword\n");
	# failure
	return undef;
    }
    # add the card to the end of the header
    push (@{$self->{CARDS}}, $card);
    # return the created card
    return $card;
}

# The get_card($keyword,$index) function retrieves the $index'th card with the
# given keyword. If $index isn't given it defaults to the first card.
sub get_card ($)
{
    # get self
    my $self = shift;
    # get parameters
    my ($keyword, $index) = @_;
    # set $index if not given
    $index = 1 unless defined ($index);
    # modify $index to deal with arrays indexing starting at zero, not one.
    --$index;
    # get 
    @cards = grep {$_->{KEYWORD} eq $keyword} @{$self->{CARDS}};
    return undef unless $#cards >= $index;
    return $cards[$index];
}

# The get_value function returns the value of a FITS card. It returns a scalar
# holding the actual value, not a FITS value object.
sub get_value ()
{
    # get self
    my $self = shift;
    # get value object (if $self is not already a value object)
    $self = $self->{VALUE} if $self->{TYPE} < $types->{UNKNOWN_TYPE};
    # return the VALUE field of the value object.
    return $self->{VALUE};
}

# The validate_keyword($keyword) takes a keyword parameter and returns true
# if the keyword is a valid FITS keyword, and false otherwise.
sub validate_keyword ($)
{
    # get the keyword parameter
    my ($keyword) = @_;
    # check and return regex
    return $keyword =~ /^[0-9A-Z_\-]{0,8}$/;
}

# The build_comment_card($value) function returns a FITS comment card with the
# given value
sub build_comment_card ($)
{
    # get parameters
    my ($value) = @_;
    # create card object
    my $card = bless ({});
    # if comment assigned string, strip quotes
    $value =~ s/^\'(.*)\'$/$1/;
    # set member variables
    $card->{TYPE}    = $types->{COMMENT_CARD};
    $card->{KEYWORD} = "COMMENT";
    $card->{VALUE}   = $value;
    # return the card
    return $card;
}

# The build_history_card($value) function returns a FITS history card with the
# given value.
sub build_history_card ($)
{
    # get parameters
    my ($value) = @_;
    # create card object
    my $card = bless ({});
    # if comment is assigned a string value, strip quotes from the string
    $value =~ s/^\'(.*)\'$/$1/;
    # set member variables
    $card->{TYPE}    = $types->{HISTORY_CARD};
    $card->{KEYWORD} = "HISTORY";
    $card->{VALUE}   = $value;
    # return the card
    return $card;
}

# The build_blank_card($value) function returns a FITS card with no keyword
# and the given value
sub build_blank_card ($)
{
    # get parameters
    my ($value) = @_;
    # create card object
    my $card = bless ({});
    # if comment assigned string, strip quotes
    $value =~ s/^\'(.*)\'$/$1/;
    # set member variables
    $card->{TYPE}  = $types->{BLANK_CARD};
    $card->{VALUE} = $value;
    # return the card
    return $card;
}

# the build_end_card function returns a FITS card with the END keyword
# (no value is associated with such cards)
sub build_end_card ()
{
    # create card object
    my $card = bless ({});
    # set card type
    $card->{TYPE} = $types->{END_CARD};
    # return the card
    return $card;
}

# the build_named_card function returns a FITS card with the given keyword,
# value, and (optional) comment.
sub build_named_card ($$;$)
{
    # get parameters
    my ($keyword, $value, $comment) = @_;
    # create card object
    my $card = bless ({});
    # set member variables
    $card->{TYPE}    = $types->{NAMED_CARD};
    $card->{KEYWORD} = $keyword;
    $card->{VALUE}   = build_value_object ($value);
    $card->{COMMENT} = $comment;
    # return card object
    return $card;
}

# the error_object function returns a FITS object whose type is ERROR. This
# is used
sub error_object ()
{
    # create and return the FITS object
    return bless ({TYPE => $types->{ERROR}});
}

# the build_value_object function takes a scalar value and returns a VALUE
# object with the appropriate TYPE field.
sub build_value_object ($)
{
    # get parameters
    my ($value) = @_;
    # declare local variables
    my ($valueObject, $string);
    # create the value object
    $valueObject = bless ({});
    # determine the type of the value and construct accordingly
    # allow double quotes or single quotes
    if ($value =~ /^\s*\'(([^\']|(?:\'\'))*)\'\s*/ or
	$value =~ /^\s*\"(([^\"]|(?:\"\"))*)\"\s*/)
	# character string
    {
	# get what's between the single quotes
	$string = $1;
	# replace any pairs of single quotes with a single single quote
	$string =~ s/\'\'/\'/g;
	# set member variables
	$valueObject->{TYPE}  = $types->{CHARACTER_STRING};
	$valueObject->{VALUE} = $string;
    }
    elsif ($value =~ /^\s*(T|F)\s*$/)
	# logical variable
    {
	# set member variables
	$valueObject->{TYPE}  = $types->{LOGICAL_VARIABLE};
	$valueObject->{VALUE} = $1;
    }
    elsif ($value =~ /^\s*((?:\+|\-)?\d+)\s*$/)
	# integer
    {
	# set member variables
	$valueObject->{TYPE}  = $types->{INTEGER};
	$valueObject->{VALUE} = $1;
    }
    elsif ($value =~ /^\s*((?:\+|\-)?\d+)\s+
	                  ((?:\+|\-)?\d+)\s*$/x)
	# complex integer
    {
	# set member variables
	$valueObject->{TYPE}  = $types->{COMPLEX_INTEGER};
	$valueObject->{VALUE} = [$1, $2];
    }
    elsif ($value =~ /^\s*((?:\+|\-)?(?:(?:\d+\.\d*)|(?:\.\d+))
			   (?:E(?:\+|\-)?\d+)?)\s*$/x)
	# real
    {
	# set member variables
	$valueObject->{TYPE}  = $types->{REAL};
	$valueObject->{VALUE} = $1;
    }
    elsif ($value =~ /^\s*((?:\+|\-)?(?:(?:\d+\.\d*)|(?:\.\d+))
			   (?:E(?:\+|\-)?\d+)?)\s+
	                  ((?:\+|\-)?(?:(?:\d+\.\d*)|(?:\.\d+))
			   (?:E(?:\+|\-)?\d+)?)\s*$/x)
	# complex real
    {
	# set member variables
	$valueObject->{TYPE}  = $types->{COMPLEX_REAL};
	$valueObject->{VALUE} = [$1, $2];
    }
    else
	# unknown
    {
	# print warning message
	warn ("WARNING - could not determine type of value \"$value\"");
	# set member variables
	$valueObject = error_object ();
	$valueObject->{VALUE} = $value;
    }
    return $valueObject;
}

# the load_from_file($filename) function reads a FITS file into the given
# FITS object.
sub load_from_file ($)
{
    # get parameters
    my ($self, $filename) = @_;
    # open file, aborting load on error
    if (not open (FITS_STREAM, "<", $filename))
    {
	# file not opened, print warning
	warn ("WARNING - Could not open FITS file at $filename");
	return undef;
    }
    # load from the opened file
    $self = $self->load_from_stream (FITS_STREAM);
    # close the file
    close (FITS_STREAM);
    return $self;
}

# the load_from_stream($stream) function reads a STREAM containing a FITS
# file and loads it into the given FITS object.
sub load_from_stream ($)
{
    # get parameters
    my ($self, $stream) = @_;
    # declare local variables
    my ($buffer, $card, $cardObj, @cards);
  readFits:
    # read 2880 byte records into buffer
    while (read ($stream, $buffer, RECORDSIZE))
    {
      readCard:
	# read 36 cards at a time (36 cards per record)
	for (1..CARDSPERRECORD)
	{
	    # get the card from the buffer
	    $card = substr ($buffer, 0, CARDIMAGELENGTH, "");
	    # parse the card
	    next readCard unless ($cardObj = parse_card ($card));
	    # terminate if at end of buffer
	    last readFits if ($cardObj->is_end_card());
	    # add card
	    push (@cards, $cardObj);
	    # (failsafe incase the read didn't get a full 2880 record).
	    last readCard if ($buffer eq "");
	}
    }
    # set member variables
    $self->{TYPE}  = $types->{PRIMARY_HEADER};
    $self->{CARDS} = \@cards;
    # return self
    return $self;
}

# the parse_card($card) function takes a string containing a FITS card and
# creates a FITS card object with the appropriate values set.
sub parse_card ($)
{
    # get parameters
    my ($card) = @_;
    # declare local variables
    my ($cardObject);
    # determine type of line
    if (length ($card) < CARDIMAGELENGTH)
	# line too short...
    {
	warn ("WARNING - Card length is too short");
	$cardObject = error_object ();
    }
    elsif ($card =~ /^\s*$/)
	# completely blank line
    {
	return undef;
    }
    elsif ($card =~ /^\s{8}/)
	# blank keyword
    {
	$cardObject = build_blank_card (substr ($card, 8));
    }
    elsif ($card =~ /^COMMENT\s/)
	# comment line
    {
	$cardObject = build_comment_card (substr ($card, 8));
    }
    elsif ($card =~ /^HISTORY\s/)
	# history line
    {
	$cardObject = build_history_card (substr ($card, 8));
    }
    else
	# keyword line
    {
	# get the keyword from the first 8 characters
	$keyword = substr ($card, 0, 8, "");
	# trim trailing spaces from keyword
	$keyword =~ s/\s*$//;
	# validate the keyword
	if (not validate_keyword ($keyword))
	{
	    warn ("WARNING - keyword $keyword invalid");
	    $cardObject = error_object ();
	}
	# check for the END keyword
	elsif ($keyword eq "END")
	{
	    $cardObject = build_end_card ();
	}
	# check for the value indicator
	elsif (substr ($card, 0, 2, "") ne "= ")
	{
	    warn ("WARNING - Value indicator for keyword $keyword ".
		  "missing");
	    $cardObject = error_object ();
	}
	# card now contains Value/Comment portion (bytes 11-80)
	else
	{
	    # split the value and comment portion of the card
	    ($value, $comment) = split_value_comment ($card);
	    # create the card object.
	    $cardObject = ($value ?
			   build_named_card ($keyword, $value, $comment):
			   error_object ());
	}
    }
    # return the card
    return $cardObject;
}

# the split_value_comment function takes a string containing a FITS value
# and comment separated by a forward slash and returns an array containing
# first the value, then the comment 
sub split_value_comment ($)
{
    # get parameters
    my ($valueComment) = @_;
    # declare local variables
    my ($value, $comment);
    # get the value and comment, if present
    if ($valueComment =~ /^\s*((?:[^\'\/]+)|(?:\'(?:(?:[^\'])|(?:\'\'))*\'))?
	                      (?:\s*\/(.*))?$/x)
    {
	($value, $comment) = ($1, $2);
	$value =~ s/\s*(.*\S)\s*/$1/;
	$comment =~ s/\s*(.*\S)\s*/$1/;
    }
    else
    {
	# string not formatted properly.
	# warn
	warn ("WARNING - couldn't parse value/comment from: \n$valueComment");
	# return error.
	return undef;
    }
    # return the value and comment
    return ($value, $comment);
}

# the write_to_file($filename) function writes a FITS object to a file.
sub write_to_file ($)
{
    # get parameters
    my ($self, $filename) = @_;
    # open file or die trying
    open ($file, ">", $filename) or
	die ("Couldn't open $filename");
    # write the object to the stream.
    $self->write_to_stream($file);
    # close the file
    close ($file);
}

# the write_to_stream($stream) function writes a FITS object to a stream.
sub write_to_stream (;$)
{
    # get parameters
    my ($self, $stream) = @_;
    # init local variables
    my ($lineCount);
    # ensure we have a header
    if (not $self->is_header ())
    {
	warn ("WARNING - only write headers to streams");
	return;
    }
    # default to STDOUT if no stream given
    $stream = STDOUT unless ($stream);
    # start line counter
    $lineCount = 0;
    # write all cards:
  write_card:
    foreach $card (@{$self->{CARDS}})
    {
	# convert the card to a string
	$line = $card->to_string();
	# skip the card if an error occured
	next write_card if ($line eq "");
	# write the line to the output stream
	print {$stream} ($line);
	# add right number of lines (some cards spill over)
	$lineCount += length ($line) / CARDIMAGELENGTH;
    }
    # add the "END" card
    print {$stream} (build_end_card ()->to_string ());
    # add to the line count
    ++$lineCount;
    # pad to full record with spaces.
    while (($lineCount % CARDSPERRECORD) != 0)
    {
	# add a card worth of spaces
	print {$stream} (" " x CARDIMAGELENGTH);
	# add one to the line count
	++$lineCount;
    }
    # handle the data portion if present
    $self->append_data ($stream);
}

# the set_data($filename, $offset, $rbytes) function sets the data portion of
# a FITS object. $filename is the source file where the data resides. $offset
# and $rbytes (defaults to 1) declare the location within the file where the
# data begins. The data is assumed to start at the $offset'th reccord, where
# the record size is defined by $rbytes. if $rbytes is not given, the offset
# is assumed to be one (the same as $rbytes being set to one.)
sub set_data ($;$$)
{
    # get parameters
    my ($self, $filename, $offset, $rbytes) = @_;
    # set DATA values
    if (-e $filename)
    {
	$self->{DATA}->{FILENAME} = $filename;
    }
    else
    {
	if ($filename =~ /^(.*\/)([^\/]*)/)
	{
	    ($path, $name) = ($1, $2);
	    print $name."\n";
	    $name = lc ($name);
	    print $name."\n";
	    $path = lc ($path) if (! -e $path.$name);
	    $filename = $path.$name;
	    print "==".$filename."\n";
	}
	else
	{
	    $filename = lc ($filename);
	}
	if (! -e $filename)
	{
	    warn ("Couldn't find file matching $filename\n");
	    return undef;
	}
    }
    $self->{DATA}->{OFFSET}   = $offset;
    $self->{DATA}->{RBYTES}   = defined ($rbytes)?$rbytes:1;
}

# the append_data($stream) function appends the DATA associated with the FITS
# header to the stream given.
sub append_data ($)
{
    # get self
    my $self = shift;
    # get stream parameter
    my ($stream) = @_;
    # declare local variables
    my ($offset, $dataSize, $dataFile, $readSize, $charsRead, $totalChars,
	$buffer);
    # append data now.
    if (exists ($self->{DATA}))
    {
	# get offset, alter for base of 0, not 1, when dealing with files
	$offset = $self->{DATA}->{OFFSET} - 1;
	# multiply by reccord size
	$offset *= $self->{DATA}->{RBYTES};
	# compute data size from FITS header.
	$dataSize = $self->compute_datasize ();
	# if data size was computable...
	if (defined $dataSize)
	{
	    # open file, defaulting to lower case on failures
	    if (not open ($dataFile, "<", $self->{DATA}->{FILENAME}))
	    {
		open ($dataFile, "<", lc ($self->{DATA}->{FILENAME})) or
		    die ("Couldn't open file: ".$self->{DATA}->{FILENAME});
		warn ($self->{DATA}->{FILENAME}." file not found, using ".
		      lc ($self->{DATA}->{FILENAME})." instead.");
	    }
	    # seek to offset in file where data begins
	    seek ($dataFile, $offset, 0);
	    # begin counting characters
	    $totalChars = 0;
	    # while more data to read
	    while ($dataSize > 0)
	    {
		# default read size to 512 bytes
		$readSize = 512;
		# decrease read size if larger than remaining data
		$readSize = $dataSize if $dataSize < 512;
		# read data
		$charsRead = read ($dataFile, $buffer, $readSize);
		# die if data ended too early
		die ("EOF found before expected")
		    if ($charsRead != $readSize);
		# write data to stream
		print {$stream} ($buffer);
		# add characters read to counter
		$totalChars += $charsRead;
		# subtract characters read from total data size
		$dataSize -= $charsRead;
	    }
	    # compute pad size for 2880 block
	    $pad = RECORDSIZE - ($totalChars % RECORDSIZE);
	    # print pad
	    print {$stream} (chr (0) x $pad);
	}
	else
	{
	    # data size not computable.
	    warn ("ERROR RETRIEVING DATA SIZE\n");
	}
    }
}

# the compute_datasize function returns the number of characters in the data
# associated with the FITS header. It is based on reserved FITS keywords.
sub compute_datasize ()
{
    # get self
    my $self = shift;
    # declare local parameters
    my (@cards, @temp, $naxis, $bitpix, $nbits, $index);
    # get card array
    @cards = @{$self->{CARDS}};
    # get bitpix card
    @temp = grep {$_->{KEYWORD} eq "BITPIX";} (@cards);
    # make sure bitpix card present
    if ($#temp != 0)
    {
	warn ("Header must contain exactly one BITPIX keyword\n");
	return undef;
    }
    # save bitpix value
    $bitpix = $temp[0]->{VALUE}->{VALUE};
    # get naxis keyword
    @temp = grep {$_->{KEYWORD} eq "NAXIS";} (@cards);
    # make sure naxis keyword present
    if ($#temp != 0)
    {
	warn ("Header must contain exactly one NAXIS keyword\n");
	return undef;
    }
    # save naxis value
    $naxis = $temp[0]->{VALUE}->{VALUE};
    if ($naxis != 0)
    {
	# if naxis is not zero, data may be present
	# set nbits to bitpix
	$nbits = abs($bitpix);
	# for each axis i
	for ($i = 1; $i <= $naxis; ++$i)
	{
	    # get the card for NAXISi
	    @temp = grep {$_->{KEYWORD} eq "NAXIS".$i;} (@cards);
	    # make sure NAXISi card exists
	    if ($#temp != 0)
	    {
		warn ("Header must contain exactly one NAXIS$i keyword\n");
		return undef;
	    }
	    # multiply nbits by axis size.
	    $nbits *= $temp[0]->{VALUE}->{VALUE};
	}
    }
    else
    {
	# naxis is zero, no data present
	$nbits = 0;
    }
    # return nbits converted from bits to bytes.
    return $nbits / 8;
}

sub to_string ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my ($string, $complexPart, $num_cards, $size);
    if ($self->is_header ())
    {
	# to_string not to be called on header (only cards or values)
	warn ("WARNING - cannot convert header to string");
	# leave string blank
	$string = undef;
    }
    elsif ($self->is_card ())
    {
	# $self is a card, check card type
	if ($self->is_named_card ())
	{
	    # normal named card, start constructing
	    # start with keyword
	    $string = $self->{KEYWORD};
	    # padd to 8 characters, using spaces, equals sign, and a space
	    $string .= (" " x (8 - length ($string)))."= ";
	    # add value as string
	    $string .= $self->{VALUE}->to_string ();
	    $skip1stReplace = FALSE;
	    # work on truncating line if too long
	    if (length ($string) > CARDIMAGELENGTH)
	    {
		# check type of value
		if ($string =~ /^(.{8})\=\s\'/)
		{
		    # value is string
		    # display truncation warning
		    warn ("Warning: fits line too long, truncating. line:\n".
		          $string."\n");
		    # save old string
		    $oldString = $string;
		    # replace overflowing text with a single quote
		    # (to close the string)
		    $string =~ s/^(.{79})(.*)$/$1\'/;
		    # add a history line
		    $string .= "HISTORY truncated line: $oldString";
		    # don't mess with the first 80 characters. They are good.
		    $skip1stReplace = TRUE;
		}
		else
		{
		    # $size represents the number of 80-character cards used
		    # start with one
		    $size = 1;
		    # while the string length is longer than the number of
		    # cards alloted
		    while (length ($string) > CARDIMAGELENGTH * $size)
		    {
			# add eight spaces after the end of the last card
			# alloted. This pushes the next region of text to
			# after the first 8 characters on the next card.
			# Any card which starts with 8 spaces can contain
			# anything thereafter.
			# Example (using 15 byte width rather than 80, for
			# readability):
			# VALUE   = 12345678901234567890
			# will become (after 3 passes)
			# VALUE   = 12345
			#         6789012
			#         3456789
			#         0
			substr ($string, CARDIMAGELENGTH * $size, 0) = " " x 8;
			# increment the number of cards
			++$size;
		    }
		    #$string = substr ($string, 0, CARDIMAGELENGTH);
		}
	    }
	    # add comment if present
	    $string .= (" " x (30 - length ($string))). " / ".$self->{COMMENT}
	      if (not $self->{COMMENT} =~ /^\s*$/);
	    # handle overflow again, due to comment
	    # start with one card
	    $size = 1;
	    # (start with two if we know the first card is allright)
	    ++$size if ($skip1stReplace);
	    # while the string is longer than the alloted cards
	    while (length ($string) > CARDIMAGELENGTH * $size)
	    {
		# see above for notes, this code is the same as the
		# previous loop
		substr ($string, CARDIMAGELENGTH * $size, 0) = " " x 8;
		++$size;
	    }
	}
	elsif ($self->is_comment_card ())
	{
	    # card is comment, start with COMMENT keyword and append the value
	    $string = "COMMENT ".$self->{VALUE};
	    # allot one card
	    $size = 1;
	    # while the string is longer than the number of alloted cards:
	    while (length ($string) > CARDIMAGELENGTH * $size)
	    {
		# see above loops which handle wrapping text for documentation.
		substr ($string, CARDIMAGELENGTH * $size, 0) = " " x 8;
		++$size;
	    }
	}
	elsif ($self->is_history_card ())
	{
	    # card is HISTORY, start with HISTORY keyword and append the value
	    $string = "HISTORY ".$self->{VALUE};
	    # allot one card
	    $size = 1;
	    while (length ($string) > CARDIMAGELENGTH * $size)
	    {
		# see above loops which handle wrapping text for documentation
		substr ($string, CARDIMAGELENGTH * $size, 0) = " " x 8;
		++$size;
	    }
	}
	elsif ($self->is_blank_card ())
	{
	    # card has no keyword (blank refers to keyword, not value).
	    # start with eight spaces, and append the value
	    $string = "        ".$self->{VALUE};
	    # allot one card
	    $size = 1;
	    while (length ($string) > CARDIMAGELENGTH * $size)
	    {
		# see above loops which handle wrapping text for documentation
		substr ($string, CARDIMAGELENGTH * $size, 0) = " " x 8;
		++$size;
	    }
	}
	elsif ($self->is_end_card ())
	{
	    # card is end card, very simple
	    $string = "END";
	}
	# figure out how many cards the string uses
	$num_cards = length ($string) / CARDIMAGELENGTH;
	# round up to whole card
	$num_cards = int (1 + $num_cards) if ($num_cards != int ($num_cards));
	# computer new size, based on integer number of cards
	$size = CARDIMAGELENGTH * $num_cards;
	# padd with spaces to that size
	$string = $string." " x ($size - length ($string));
    }
    elsif ($self->is_value ())
    {
	# object is value
	if ($self->is_character_string ())
	{
	    # value is string
	    # get string value
	    $string = $self->{VALUE};
	    # replace single quotes with two adjacent single quotes
	    $string =~ s/\'/\'\'/g;
	    # padd on right up to eight characters
	    $string .= " " x (8 - length ($string));
	    # wrap in single quotes
	    $string = "'$string'";
	}
	elsif ($self->is_logical_variable ())
	{
	    # value is logical (true or false)
	    # right-justify value to 20 characters
	    $string = (" " x 19).$self->{VALUE};
	}
	elsif ($self->is_integer () or
	       $self->is_real ())
	{
	    # value is non-complex number
	    # get value
	    $string = $self->{VALUE};
	    # right justify to 20 characters
	    $string = (" " x (20 - length ($string))).$string;
	}
	elsif ($self->is_complex_integer () or
	       $self->is_complex_real ())
	{
	    # value is complex integer or real (real refers to floating point
	    # not real numbers verses rational/irrational/complex)
	    # get non-complex part
	    $string = $self->{VALUE}->[0];
	    # right justify to 20 characters
	    $string = (" " x (20 - length ($string))).$string;
	    # get complex part
	    $complexPart = $self->{VALUE}->[1];
	    # right justify to 20 characters
	    $complexPart = (" " x (20 - length ($complexPart))).$complexPart;
	    # append complex part to non-complex part
	    $string .= $complexPart;
	}
	else
	{
	    # unknown value, display warning
	    warn ("WARNING - to_string called on unknown value type");
	    # leave $string blank
	    $string = undef;
	}
    }
    # return the constructed string
    return $string;
}

# Object Type Determiners

# the get_type_string function returns a string representation of the type
# of value stored in a value object.
sub get_type_string ()
{
    # get self
    my ($self) = @_;
    # grep the types hash for a matching key
    my @results = grep {$types->{$_} == $self->{TYPE}} (keys (%{$types}));
    # return the key if it exists
    return $results[0] if ($#results == 0);
    # otherwise... return undef
    return undef;
}

# the is_header function returns true if it is called on a FITS header object
sub is_header ()
{
    # get self
    my ($self) = @_;
    # return type check
    return ($self->{TYPE} == $types->{PRIMARY_HEADER} or
	    $self->{TYPE} == $types->{EXTENDED_HEADER});
}

# the is_primary_header function returns true if it is called on a FITS
# primary header object
sub is_primary_header ()
{
    # get self
    my ($self) = @_;
    # return true if primary header
    return $self->{TYPE} == $types->{PRIMARY_HEADER};
}

# the is_extended_header function returns true if it is called on a FITS
# extended header
sub is_extended_header ()
{
    # get self
    my ($self) = @_;
    # return true if extended header
    return $self->{TYPE} == $types->{EXTENDED_HEADER};
}

# the is_card function returns true if it is called on a FITS card object.
sub is_card ()
{
    # get self
    my ($self) = @_;
    # check for any kind of card
    return ($self->{TYPE} == $types->{NAMED_CARD} or
	    $self->{TYPE} == $types->{END_CARD} or
	    $self->{TYPE} == $types->{COMMENT_CARD} or
	    $self->{TYPE} == $types->{HISTORY_CARD} or
	    $self->{TYPE} == $types->{BLANK_CARD});
}

# the is_named_card function returns true if it is called on a FITS named card
# object
sub is_named_card ()
{
    # get self
    my ($self) = @_;
    # return true if named card
    return $self->{TYPE} == $types->{NAMED_CARD};
}

# the is_end_card function returns true if it is called on an FITS end card
# object
sub is_end_card ()
{
    # get self
    my ($self) = @_;
    # return true if end card
    return $self->{TYPE} == $types->{END_CARD};
}

# the is_comment_card function returns true if it is called on a FITS comment
# card object
sub is_comment_card ()
{
    # get self
    my ($self) = @_;
    # return true if comment card
    return $self->{TYPE} == $types->{COMMENT_CARD};
}

# the is_history_card function returns true if it is called on a FITS history
# card object
sub is_history_card ()
{
    # get self
    my ($self) = @_;
    # return true if history card
    return $self->{TYPE} == $types->{HISTORY_CARD};
}

# the is_blank_card function returns true if it is called on a FITS card with
# no keyword.
sub is_blank_card ()
{
    # get self
    my ($self) = @_;
    # return true if blank card
    return $self->{TYPE} == $types->{BLANK_CARD};
}

# the is_value function returns true if it is called on a FITS value object.
sub is_value ()
{
    # get self
    my ($self) = @_;
    # check for any type of value object
    return ($self->{TYPE} == $types->{CHARACTER_STRING} or
	    $self->{TYPE} == $types->{LOGICAL_VARIABLE} or
	    $self->{TYPE} == $types->{INTEGER} or
	    $self->{TYPE} == $types->{COMPLEX_INTEGER} or
	    $self->{TYPE} == $types->{REAL} or
	    $self->{TYPE} == $types->{COMPLEX_REAL});
}

# the is_character_string function returns true if it is called on a FITS
# value which is a character string
sub is_character_string ()
{
    # get self
    my ($self) = @_;
    # return true if character string
    return $self->{TYPE} == $types->{CHARACTER_STRING};
}

# the is_logical_variable function returns true if it is called on a FITS
# value which is a logical variable (T or F)
sub is_logical_variable ()
{
    # get self
    my ($self) = @_;
    # return true if logical variable
    return $self->{TYPE} == $types->{LOGICAL_VARIABLE};
}

# the is_integer function returns true if it is called on a FITS value which
# is an integer
sub is_integer ()
{
    # get self
    my ($self) = @_;
    # return true if integer
    return $self->{TYPE} == $types->{INTEGER};
}

# the is_complex_integer function returns true if it is called on a FITS value
# which is a complex integer
sub is_complex_integer ()
{
    # get self
    my ($self) = @_;
    # return true if complex integer
    return $self->{TYPE} == $types->{COMPLEX_INTEGER};
}

# the is_real function returns true if it is called on a FITS value object
# which is a real number.
sub is_real ()
{
    # get self
    my ($self) = @_;
    # return true if real
    return $self->{TYPE} == $types->{REAL};
}

# the is_complex_real function returns true if it is called on a FITS value
# object which is a complex real number
sub is_complex_real ()
{
    # get self
    my ($self) = @_;
    # return true if complex real
    return $self->{TYPE} == $types->{COMPLEX_REAL};
}

# the is_numeric_value function returns true if it is called on a FITS value
# object which is any kind of number
sub is_numeric_value ()
{
    # get self
    my ($self) = @_;
    # return true if any kind of number.
    return ($self->{TYPE} == $types->{INTEGER} or
	    $self->{TYPE} == $types->{COMPLEX_INTEGER} or
	    $self->{TYPE} == $types->{REAL} or
	    $self->{TYPE} == $types->{COMPLEX_REAL});
}

# package must return true
1;
