/*
  Program: break_fitstab

  Routine to prepare a FITS ASCII Table file for processing by 'fits2pds'.
  This program does the following:

      1. Separates the FITS header blocks from the data.
      2. Inserts PDS carriage control into the data file and deletes
         blank lines
      3. Modifies the FITS EXTENSION header NAXIS1 value to reflect the
         record length with delimiters (\r\n)
      4. Writes the modified data into a separate file.

  Format:

     % break_fitstab <input file> <output header> <output data>

  14 July 1995, acr:
  08 Jan 1998, acr: Fixed bug in read/write loop which caused a spurious
                    error messages in files where NAXIS1xNAXIS2 is an even
                    multiple of 2880

*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#define BLOCKSIZE 2880

/*=========================================================================*/

main (argc,argv)
  int    argc;
  char  *argv[];

{ FILE *ifp,*hfp,*dfp;    /* File pointers */

  char  hdr[36][80];      /* header block buffer */
  char  block[BLOCKSIZE]; /* data block buffer */

  int   i,j,k;            /* loop/subscript */
  int   bytes;            /* i/o byte count */
  int   done;             /* logical flag */
  int   naxis1;           /* number of bytes per record */
  int   naxis2;           /* number of records */
  int   bc;

  /*-----------------------------------------------------------------------*/
  /* Get the input arguments and open the files:                           */

  if (argc != 4)
    { fprintf (stderr,"Usage: break_fitstab <input> <output label> ");
      fprintf (stderr,"<output data>\n");
      exit(0);
    }

  if ((ifp=fopen(argv[1],"rb")) == NULL)
    { fprintf (stderr,"break_fitstab: Unable to open %s for reading (%s).\n",
                      argv[1],strerror(errno));
      exit(10);
    }

  if ((hfp=fopen(argv[2],"wb")) == NULL)
    { fprintf (stderr,"break_fitstab: Unable to open %s for writing (%s).\n",
                      argv[2],strerror(errno));
      exit(10);
    }

  if ((dfp=fopen(argv[3],"w")) == NULL)
    { fprintf (stderr,"break_fitstab: Unable to open %s for writing (%s).\n",
                      argv[3],strerror(errno));
      exit(10);
    }

  /*-----------------------------------------------------------------------*/
  /* Step through the header blocks, looking for the XTENSION header:      */

  done = -1;
  while (done != 1)

    /* We assume that there is only one XTENSION header in the file - the 
       TABLE extension.  This loop is stopped when an "END" record is found
       after the first extension header block.
    */

    { if (fread((void *)hdr,BLOCKSIZE,1,ifp) != 1)
        { fprintf (stderr,"break_fitstab: Error reading from %s (%s).\n",
                          argv[1],strerror(errno));
          exit(99);
        }

      if (strncmp(hdr[0],"XTENSION",8) == 0)   /* Found extension header */
        { if (strncmp(hdr[0]+11,"TABLE",5) != 0)  /* Make sure its OK */
            { fprintf (stderr, "break_fitstab:  Wait a minute - ");
              fprintf (stderr, "this XTENSION isn't a TABLE!\n");
              fprintf (stderr, "I can't handle complex FITS files...\n");
              exit(100);
            }

          /* Now we know it's a TABLE, so the fourth line must contain
             NAXIS1:
          */

          if (strncmp(hdr[3],"NAXIS1",6) != 0)
            { fprintf (stderr, "break_fitstab: NAXIS1 wasn't found where ");
              fprintf (stderr, "it should have been in the XTENSION header.\n");
              exit(101);
            }

          /* Read the NAXIS1 value in columns 11-30, add 2, and write it
             back:
          */

          naxis1 = atoi(hdr[3]+10) + 2;
          j = hdr[3][30];   /* save this value - it's about to get squashed */
          sprintf (hdr[3]+10,"%20d",naxis1);
          hdr[3][30] = j;

          /* Read the NAXIS2 value in the next record: */

          if (strncmp(hdr[4],"NAXIS2",6) != 0)
            { fprintf (stderr, "break_fitstab: NAXIS2 wasn't found where ");
              fprintf (stderr, "it should have been in the XTENSION header.\n");
              exit(102);
            }

          naxis2 = atoi(hdr[4]+10);

          /* We've encountered an XTENSION block, so set 'done': */

          done = 0;

        }

      /* And write the block out: */

      if (fwrite((void *)hdr,BLOCKSIZE,1,hfp) != 1)
        { fprintf (stderr, "break_fitstab: Error on output to %s (%s).\n",
                            argv[2],strerror(errno));
          exit(102);
        }

      /* Now search the block for an "END" record, but only if we've already
         processed the extension header (done was initialized to -1):
      */

      if (!done)
        { for (i=0; i<36; ++i)
            { if (strncmp(hdr[i],"END       ",10) == 0) done = 1; }
        }
    }

    /* Close the header file: */

    fclose(hfp);

  /*-----------------------------------------------------------------------*/
  /* Now process the data part one byte at a time:                         */

  /* Note that the file is assumed to contain an integral number of FITS
     blocks.
  */

  if (fread((void *)block,1,BLOCKSIZE,ifp) != BLOCKSIZE)
    { fprintf (stderr,"break_fitstab: Error reading %s (%s).\n",
                       argv[1],strerror(errno));
      exit(200);
    }
  k = 0;

  for (i=0; i<naxis2; ++i)          /* for each record */
    { for (j=0; j<naxis1-2; ++j)    /* for each byte - but there's no CC yet  */
        { fputc(block[k++],dfp);

          /* If we've run out of data in this block but still have more to go,
             read the next block:
          */

          if (k == BLOCKSIZE  &&  (i < naxis2-1  ||  j < naxis1-3))
            { if ((bc=fread((void *)block,1,BLOCKSIZE,ifp)) != BLOCKSIZE)
                { fprintf (stderr,"break_fitstab: Error 2 reading %s (%s).\n",
                                  argv[1],strerror(errno));
                  fprintf (stderr," Only read %d bytes in last block.\n",bc);
                  exit(201);
                }
              k = 0;
            }
        }

      /* Write the carriage-control characters: */

      fputc('\r',dfp);
      fputc('\n',dfp);
    }

  /* Done with the data file: */

  fclose(dfp);
  exit(0);
}

           
