/*
  brkstrm:  Breaks a stream-format file into one with regular, fixed-length
            records delimited by newline (\n) characters.  The default break
            length is 80 (output length = 81, with the \n).  The number of
            output lines is reported unless the output is going to stdout.

  Format:  % brkstrm  [-d delimiter] [-r length] [ input_file [ output_file ] ]

             -d values:  0   Unix (\n)
                         1   Mac  (\r)
                         2   PC   (\r\n)

  09 May 1994, acr:  Replaces 'fitsform' routine.
  13 May 1994, acr:  supressed line count reporting for stdout
  12 Oct 1994, acr:  Fixed output so that null characters are not lost
  11 Jun 1996, acr:  Removed input record length limitation; added delimiter
                     option

*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#define  DEFLENGTH     80     /* Default input record length   */
#define  DEFDELIMIT     0     /* Default delimiter type (Unix) */


main (argc,argv)
  int   argc;
  char *argv[];

{ FILE     *ifp, *ofp;         /* file pointers  */
  char     *line;              /* input/output buffer */
  int       length;            /* input (read) line length */
  int       olen;              /* output (write) line length */
  int       count;             /* output line count */
  int       i,j,k;             /* loop/subscript */
  char     *rem;               /* conversion error checker */
  int       delimit;           /* delimiter flag: 0 = \n; 1 = \r; 2 = \r\n */

  length = DEFLENGTH;    /* Set default record length */
  delimit = DEFDELIMIT;  /* Set default delimiter type */
  count = 0;             /* Initialize record count   */
  ifp=stdin;             /* Use stdio unless otherwise specified */
  ofp=stdout;

  /* Gather up the arguments and open files as needed: */

  i = 1;
  while (i < argc)
    { if (strcmp(argv[i],"-r")==0)
        { i++;
          length = strtol(argv[i],&rem,10);
          if ( strcmp(rem,"")  ||  length < 1 )
            { fprintf (stderr,"brkstrm: Invalid record length, %s.\n",argv[i]);
              exit(10);
            }
        }

      else if (strcmp(argv[i],"-d")==0)
        { i++;
          if (strcmp(argv[i],"unix")==0)        /* '\n' only */
            { delimit = 0; }
          else if (strcmp(argv[i],"mac")==0)    /* '\r' only */
            { delimit = 1; }
          else if (strcmp(argv[i],"pc")==0)     /* "\r\n"    */
            { delimit = 2; }
          else
            { fprintf (stderr,"brkstrm: Invalid record delimiter type, %s.\n",
                               argv[i]);
              exit(11);
            }
        }

      else if (ifp==stdin)
        { if ((ifp=fopen(argv[i],"r"))==stdin)
            { fprintf (stderr,"brkstrm: Unable to open %s for reading (%s)\n",
                       argv[i],strerror(errno));
              exit(20);
            }
        }

      else if (ofp==stdout)
        { if ((ofp=fopen(argv[i],"w"))==stdout)
            { fprintf (stderr,"brkstrm: Unable to open %s for writing (%s)\n",
                        argv[i],strerror(errno));
              exit(21);
            }
        }

      else
        { fprintf (stderr,
              "Usage: brkstrm [-d delimiter] [-r length] [input [output]]\n");
          exit(0);
        }

      i++;
    }


/*---------------------------------------------------------------------------*/

  /* Allocate an input line buffer with room for the delimiters and string
     terminator; set the ouput write length:
  */

  line = (char *)malloc(length+3);

  /* Set the end of record characters (these are never overwritten): */

  if (delimit==0)       /* Unix */
    { line[length]   = '\n';
      line[length+1] = '\0';
      olen = length+1;
    }
  else if (delimit==1)  /* Mac  */
    { line[length]   = '\r';
      line[length+1] = '\0';
      olen = length+1;
    }
  else                  /* PC   */
    { line[length]   = '\r';
      line[length+1] = '\n';
      line[length+2] = '\0';
      olen = length+2;
    }
  
  /* Loop through the characters in the file, getting 'length' of them each 
     time:
  */

  while (!feof(ifp))
    { if ((i=fread((void *)line,1,length,ifp)) > 0)

      /* This tries to read 'length' bytes from the input file.  If it reads
         fewer, assume this is because of end of file.  If there were any
         bytes read, pad the record out to the given record length and print
         it anyway.  If no bytes were read, end without writing a record:
      */

        { if (i < length)
            { for (j=i; j<length; ++j) line[j] = ' '; }
          fwrite((void *)line,olen,1,ofp);
          count++;
        }
    }

  /* Close files, report and exit: */

  if (ifp != stdin)  fclose(ifp);
  if (ofp != stdout) fclose(ofp);

  if (ofp != stdout) fprintf (stdout,"brkstrm: %d records written.\n",count);
  exit(0);
}
