#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <string.h>
#include <errno.h>

#define  TRUE  1
#define  FALSE 0

#define  BYTE   0
#define  I2     1
#define  I4     2
#define  R4     3
#define  R8     4

/* fits_datachk

   Format: % fits_datachk [-n <null_value>] <input_file>

   This program runs a few simple data checks on a simple FITS image (i.e.,
   just a header and a primary data unit).  It reads the header to determine
   data type and dimentions, then finds the end of the header and checks that
   it is padded with blanks.

   Then it reads the data one pixel at a time, finding the extrema. It will
   also check past the end of the data to ensure that the data segment is 
   padded with nulls.

   The BLANK, DATAMAX, DATAMIN, BSCALE and BZERO keywords are also collected
   from the label and checked against the data values in the array.  BLANK  
   will be checked for real data as well, although this does not conform to
   the FITS standard.  

   The user may enter an effective BLANK value via the "-n" option.

   Output is to stdout.

   21 Oct 2003, ACR.

*/
   
main (int argc, char *argv[])
  { FILE     *ifp;
    int       i,j,k;             /* loop/subscript     */
    char     *key,*val;          /* for parsing FITS header lines */
    char      line[81];          /* header line buffer */
    char      buff[2880];        /* input block buffer */
    int       databytes;         /* Elements size in bytes */
    int       int_data;          /* TRUE for integer data */
    double    user_r8blank;      /* for "-n" values entered */
    char      infile[250];       /* Input file name */
    char     *databuff;          /* pointer to binary data buffer */
    long int  databuff_size;     /* Total bytes in the input data buffer */
    long int  databuff_elements; /* Number of elements in the buffer */
    long int  bytes_read;

    /* Value buffers and Extrema: */

    double    realmax, realmin; 
    double    infilemax, infilemin, r8value;

    unsigned char ub;
    short    int  i2;
    long     int  i4;
    float         r4;
    double        r8;
    
    /* Header keywords: */

    double          bzero, bscale, r8blank;
    double          datamax, datamin;
    int             bitpix;
    int             naxis;
    int             naxisn[1000];

    /* Counters and flags: */

    int      user_blank;
    int      not_done;
    int      datamin_found, datamax_found;
    int      blank_found;
    int      data_type;
    long int lt_realmin, gt_realmax, eq_blank, eq_userblank;
    int      nonnull_found;

    /*===================================================================*/

    /* Collect arguments and open input file: */

    if (argc != 2  &&  argc != 4)
      { fprintf (stderr, 
                 "Usage: fits_datachk [-n <null_value>] <input_file>\n");
        exit(0);
      }

    user_blank = FALSE;

    i = 1;
    while (i < argc)
      { if (strcmp(argv[i],"-n") == 0)
          { i++;
            user_r8blank = atof(argv[i]);
            user_blank = TRUE;
          }

        /* Better be the input file and the last thing on the input line: */

        else
          { if (i != (argc-1))
              { fprintf (stderr, 
                     "Usage: fits_datachk [-n <null_value>] <input_file>\n");
                exit(0);
              }

            strcpy(infile,argv[i]);
            if ((ifp = fopen(argv[i],"rb")) == NULL)
              { fprintf (stderr,
                     "Unable to open '%s' for reading (%s).\n",
                     argv[i], strerror(errno));
                exit(100);
              }
          }
        i++;
      }

    /*--------------------------------------------------------------------*/
    /* Reading the FITS header: */

    line[80] = '\0';
    not_done = TRUE;

    datamin_found = FALSE;
    datamax_found = FALSE;
    blank_found   = FALSE;
    bzero  = 0.0;
    bscale = 1.0;
    naxis  = 0;

    while (not_done)
      { fread((void *)buff,2880,1,ifp);

        /* Loop through the records, looking for keywords of interest: */

        for (i=0; i<36; i++)
          { memcpy((void *)line,(void *)(buff + (i*80)), 80);

            /* We parse out the keyword and value.  If either is missing,
               we go to the next line.  When we find the "END" keyword, 
               we're done:
            */

            key = strtok(line," =");
            if (!key) continue;
            if (strcmp(key,"END") == 0)
              { not_done = FALSE; 
                break;
              }

            val = strtok(NULL," ");
            if (!val) continue;

            if (strcmp(val,"=") == 0) val = strtok(NULL," ");
            if (!val) continue;
 
            if (strcmp(key,"BITPIX") == 0)
              { bitpix = (int)atoi(val); }

            else if (strcmp(key,"NAXIS") == 0)
              { naxis  = (int)atoi(val); }

            else if (strncmp(key,"NAXIS",5) == 0)
              { j = (int)atoi(key+5);
                naxisn[j] = (int)atoi(val);
              }

            else if (strcmp(key,"BSCALE") == 0)
              { bscale = atof(val); }

            else if (strcmp(key,"BZERO") == 0)
              { bzero  = atof(val); }

            else if (strcmp(key,"BLANK") == 0)
              { r8blank = atof(val); 
                blank_found = TRUE;
              }

            else if (strcmp(key,"DATAMAX") == 0)
              { datamax = atof(val);
                datamax_found = TRUE;
              }

            else if (strcmp(key,"DATAMIN") == 0)
              { datamin = atof(val);
                datamin_found = TRUE;
              }

          }

      }

    /* Check the last block for non-blanks following the "END" keyword:

    check_end_of_header(buff);

    /*--------------------------------------------------------------------*/
    /* Data Parameters */

    /* First, determine the data type: */

    if (bitpix < 0)
      { int_data  = FALSE;
        databytes = -bitpix/8;
      }
    else
      { int_data  = TRUE;
        databytes = bitpix/8;
      }

    /* Set the data_type flag and initialize the relevant extrema: */

    if (int_data  &&  databytes == 1)
      { data_type = BYTE; }
    else if (int_data  &&  databytes == 2)
      { data_type = I2; }
    else if (int_data  &&  databytes == 4)
      { data_type = I4; }
    else if (databytes == 4)
      { data_type = R4; }
    else if (databytes == 8)
      { data_type = R8; }
    else
      { fprintf (stderr,"Unrecognized data type: BITPIX %d\n",bitpix);
        fprintf (stderr,"PROCESSING STOPPED\n");
        exit(200);
      }

    /* Now we need the data buffer.  We'll use the total size of the
       1-(naxis-1) axes and allocate a void space:
    */

    databuff_elements = 1;

    for (i=1; i<naxis; i++)
      { databuff_elements = databuff_elements * naxisn[i]; }

    databuff_size = databuff_elements * databytes;

    /* Do a quick sanity check: */

    if (databuff_size == 0  ||  naxisn[naxis] == 0  ||  naxis == 0)
      { fprintf (stderr,"Header axis values indicate no data.\n");
        fprintf (stderr,"PROCESSING STOPPED\n");
        exit(0);
      }

    databuff = malloc(databuff_size);


    /* If the header contained DATAMAX or DATAMIN, we do a quick calculation
       here to convert this to the corresponding actual file value, to save
       time in later comparisons:
    */

    if (datamax_found)
      { realmax = (datamax - bzero) / bscale;
        realmin = (datamin - bzero) / bscale;
      }

        /* ...NOTE that in the above, I'm using "real" to mean the value
                   that exists in the file, because the FITS standard
                   defines DATAMAX/MIN to be the (real) physical value.
                   In this program, "real" is what is read directly from 
                   the bits.
        */

    /*--------------------------------------------------------------------*/
    /* Data Loop */

    lt_realmin = 0;
    gt_realmax = 0;
    infilemin  =  DBL_MAX;
    infilemax  = -DBL_MAX;

    /* We'll loop through the data segment in step sizes = databuff_size: */

    for (j=0; j<naxisn[naxis]; j++)
      { fread((void *)databuff,1,databuff_size,ifp);

        /* Check each element for extrema.  We convert each element to 
           'double' for comparison purposes:
        */

        for (i=0; i<databuff_elements; i++)
          { switch (data_type)
              { case BYTE : 
                    memcpy((void *)&ub,(void *)(databuff+i),1);
                    r8value = (double)ub;
                    break;
                case I2   : 
                    memcpy((void *)&i2,(void *)(databuff + (i*2)),2);
                    r8value = (double)i2; 
                    break;
                case I4   : 
                    memcpy((void *)&i4,(void *)(databuff + (i*4)),4);
                    r8value = (double)i4; 
                    break;
                case R4   : 
                    memcpy((void *)&r4,(void *)(databuff + (i*4)),4);
                    r8value = (double)r4; 
                    break;
                case R8   : 
                    memcpy((void *)&r8,(void *)(databuff + (i*8)),8);
                    r8value = r8; 
                    break;
                default   : break;
              }

            if (blank_found  &&  r8value == r8blank)
              { eq_blank++; }
            else if (user_blank  &&  r8value == user_r8blank)
              { eq_userblank++; }
            else
              { infilemax = (r8value > infilemax)? r8value : infilemax;
                infilemin = (r8value < infilemin)? r8value : infilemin;

                if (datamin_found  &&  r8value < realmin) lt_realmin++;
                if (datamax_found  &&  r8value > realmax) gt_realmax++;
              }
          }
      }

    /* From this point on, everything should be null, so we'll read to the
       end of file checking that in increments of the element size:
    */

    nonnull_found = 0;
    bytes_read=fread((void *)databuff,1,databuff_size,ifp);

    while (bytes_read > 0)
      { for (i=0; i<bytes_read; i++)
          { if ((char)databuff[i] != '\0') nonnull_found++; }
        bytes_read=fread((void *)databuff,1,databuff_size,ifp);
      }

    /* Close up the input file: */

    close(ifp);

    /*--------------------------------------------------------------------*/
    /* Report */

    fprintf (stdout,"\n%s:\n", infile);

    fprintf (stdout,"\n");

    fprintf (stdout,"   Maximum value in file: ");
    switch (data_type)
      { case BYTE : fprintf(stdout,"%3d\n",(int)infilemax); break;
        case I2   : fprintf(stdout,"%6d\n",(int)infilemax); break;
        case I4   : fprintf(stdout,"%15d\n",(int)infilemax); break;
        case R4   : fprintf(stdout,"%10.5f\n",infilemax); break;
        case R8   : fprintf(stdout,"%20.10f\n",infilemax); break;
        default   : fprintf(stdout,"---\n");
      }

    fprintf (stdout,"   Minimum value in file: ");
    switch (data_type)
      { case BYTE : fprintf(stdout,"%3d\n",(int)infilemin); break;
        case I2   : fprintf(stdout,"%6d\n",(int)infilemin); break;
        case I4   : fprintf(stdout,"%15d\n",(int)infilemin); break;
        case R4   : fprintf(stdout,"%10.5f\n",infilemin); break;
        case R8   : fprintf(stdout,"%20.10f\n",infilemin); break;
        default   : fprintf(stdout,"---\n");
      }

    if (datamax_found)
      { fprintf (stdout, 
                    "   DATAMAX = %f,  Number of values > DATAMAX = %d\n",
                         datamax, gt_realmax);
      }
    if (datamin_found)
      { fprintf (stdout, 
                    "   DATAMIN = %f,  Number of values < DATAMIN = %d\n",
                         datamin, lt_realmin);
      }


    if (blank_found)
      { fprintf (stdout, 
                    "   BLANK   = %f,  Number of BLANK values = %d\n",
                         r8blank, eq_blank);
      }
    if (user_blank)
      { fprintf (stdout, 
                    "   -n Null = %f,  Number of null values = %d\n",
                         user_r8blank, eq_userblank);
      }

    fprintf (stdout,"\n");

    if (nonnull_found)
      { fprintf (stdout, "Found %d non-null bytes after the end of data.\n"); }
    else
      { fprintf (stdout, "Data are null-padded correctly.\n"); }

    fprintf (stdout,"\n");


    /* All done: */

    exit(0);

  }

/*_________________________________________________________________________*/
/*_________________________________________________________________________*/
/*_________________________________________________________________________*/

void check_end_of_header(char *block)

  /* Routine to check the blank padding following the "END" statement of
     the header.  Header blocks should be padded with blanks only and
     the "END" statement should be by itself on an otherwise blank line.
  */

  { char line[81];   

    int  i,j,k;
    int  padding_ok;
    int  found_end;

    found_end  = FALSE;
    padding_ok = TRUE;
    line[80]   = '\0';

    /* We'll loop backward from the end of the block: */

    for (i=35; i>=0; i--)
      { memcpy((void *)line,(void *)(block + (i*80)),80);
        if (strncmp(line,"END ",4) == 0)
          { found_end = TRUE;

            /* Check the blank padding on this line: */

            if (strspn((line+3)," ") != 77)
              { padding_ok = FALSE; }

            /* We can break out of this loop as soon as we find the END: */

            break;
          }

        /* If we didn't find the END, check for all blank lines: */

        else
          { if (strspn(line," ") != 80)
              { padding_ok = FALSE; }
          }
      }

    /* If we found the END, check the padding and print a warning if needed: */

    if (found_end  &&  !padding_ok)
      { fprintf (stderr,"WARNING: Non-blanks following the END statement");
        fprintf (stderr," in the header.\n");
      }

    /* All done: */

    return;
  }
