#!/usr/bin/perl 

use Getopt::Std;
#  
#  Programmer:  Dorian Winterfeld
#  Date:        8/21/95
#  Revisions:   v1.0:  11/12/95 fixed a bug
#		v1.1:  96/7/17ij  Scans *all* files in a directory structure
#			now--not just ones with certain dot-extensions.
#  Description: Routine to generate an index file from FITS files.
#
#  08 Aug 1996, acr: re-wrote heart of routine following pdsidx as model. 
#                    added the -a and -r options.
#  07 Oct 1996, acr: fixed bug that truncated strings with the comment delimiter
#                    (/) embedded in them
#  03 Apr 1997, acr: fixed dumb bug that short-circuited directory recursion
#  26 Sep 2006, acr: Modified to use Getopt::Std;
#
#  Input options:
#
#     -a            Check all files for FITS headers
#     -n            Do not use default keywords
#     -r            Descend recursively through subdirectories
#     -o <file>     Send output to <file>
#     -f <file>     Read keywords from <file>
#     -d <dir>      Start at directory <dir>
#
#
#  Define the list of valid special fields as an associative array for easy
#  searching:

%SPECIAL = ('^PATH', 1, '^FILE', 1);

#
#  Start by dealing with the command line arguments and collecting the various
#  flags, fields, and file names.  Getopts collects the options and their
#  arguments, adjusting the ARGV array accordingly.
#  

getopts('anro:f:d:');

#
#  Formatting and keyword selection are taken care of through two arrays:
#  one associative, the other a straight array.  The associative array is
#  indexed on keyword name (including path) and returns the number of bytes
#  for the output field width.  The second contains a list of the keywords
#  to be indexed, in the order in which they appeared on the command line
#  and/or in the field list file (default fields come first).  This is used
#  to output the fields in order.
#
#
#  If the "-n" switch was used, then no default fields are defined.  
#  Otherwise, start with the default field list:
#

if (!$opt_n)
  { $num_fields = 4;
    %size = ( '^PATH', 25, '^FILE',17, 'BITPIX', 3, 'NAXIS', 3 );
    @name = ( '^PATH', '^FILE', 'BITPIX','NAXIS' );
  }

#
#  Look for an explicit output file and open it:
#

if ($opt_o)
  { open (OUTPUT,">$opt_o") || die "Could not open '$opt_o' for output.\n"; }

#
#  Next, check to see if a file of field descriptions was included.  If so,
#  read it:
#

if ($opt_f)
  { open (FIELDS,$opt_f)  ||  die "Could not open '$opt_f'\n";

    while ($line=<FIELDS>)
      { chop $line;
        if ( $line =~/^\s*$/  ||  $line =~ /^\s*#.*$/ )
          { next; }

        $line =~ s/#.*$//;                     # Delete trailing commentse
        ($field,$width) = split(/\//,$line);   # Split into fields
        if ($width eq "") { $width = 10; }     # insert default width

        $tmpsize[$num_fields] = $width;
        $name[$num_fields] = $field;
        $num_fields++;
      }
    close(FIELDS);
  }

#
#  Finally, retrieve field names from the command line, if any:
#

for ($i=0; $i<=$#ARGV; $i++)     # Get field descriptions
  { ($fname, $fwidth) = split(/\//,$ARGV[$i]);

    if ($fwidth eq "")
      { $fwidth = 10; }
    if ($fwidth < 1)
      { printf STDOUT "fitsidx: Invalid field width (%s) ignored.\n",$fwidth;
        $fwidth = 10;
      }
    $tmpsize[$num_fields] = $fwidth;
    $name[$num_fields] = $fname;
    $num_fields++;
  }

#
#  Format and do syntax-checking for all collected field names:
#

$num_fields = &checknames($num_fields,*name,*tmpsize);
if ($num_fields < 1)
  { printf STDOUT "fitsidx: no valid fields entered.\n";
    exit;
  }

#
#  And store output field sizes in an associative array:
#

for ($i=0; $i<$num_fields; $i++)
  { if ($tmpsize[$i] != 0)     #  Special Fields may have no size in the array
      { $size{$name[$i]} = $tmpsize[$i]; }
  }


#for ($i=0; $i<$num_fields; $i++)
#  { printf "Keyword: %-25.25s  Width: %2d\n", $name[$i], $size{$name[$i]};
#  }

#
#  Determine the starting directory:
#

if ($opt_d)
  { if (-d $opt_d)        # Make sure it's a directory
      { $start_dir = $opt_d; }
    else
      { printf STDOUT "fitsidx: Invalid root directory (%s).\n", $opt_d;
        exit;
      }
  }
else
  { $start_dir = '.'; }

#
#  Now call the recursive routine which will process this and all
#  subdirectories, if requested:
#

&runfiles($start_dir,$opt_a,$opt_r);

#=============================================================================

sub checknames
  { local ($n,*fldnm,*width) = @_;
    local ($name,$i,$j,$k,$count,@tmpname,@tmpsize);

    # First, run through the names, do syntax-checking and put in
    # uniform format:

    foreach $name (@fldnm)
      {  
         # Delete blanks:
 
         $name =~ s/  *//g;

         # Force to upper case:

         $name =~ tr/a-z/A-Z/;

         # Insert "[1]" before unnumbered extensions:

         $name =~ s/([A-Z]+)\./\1\[1\]./g;

         # Check for equal numbers of []:
         if (split(/\[/,$name) != split(/\]/,$name))
           { printf STDOUT "fitsidx: Invalid field name, \"%s\".\n",$name;
             $name = "";
             next;
           }

         # If there are brackets, make sure there are only digits 
         # between them:

         if ( ($name =~ /\[/)  &&  ($name =~ /\[\d*[^0-9\]]\d*\]/) )
           { printf STDOUT "fitsidx: Invalid field name, \"%s\".\n", $name;
             $name = "";
             next;
           }

         # Check for special fields (beginning with '^'):

         if ($name =~ /^\^\w+/)
           { if (!$SPECIAL{$name})     # Is it on the Special Field list?
               { printf STDOUT "fitsidx: Invalid field name, \"%s\".\n",$name;
                 $name = "";
               }
             next;    # No more checks for special names
           }

         # Finally, make sure there are only valid characters, and it
         # begins with a letter:

         if ($name !~ /^[A-Z][A-Z_0-9\-\[\]\.]*[^\.]$/)
           { printf STDOUT "fitsidx: Invalid field name, \"%s\".\n", $name;
             $name = "";
             next;
           }
      }

    #  Now, step through the lists and remove null field names:

    $count = 0;
    for ($i=0; $i<$n; $i++)
      { if ($fldnm[$i] ne "")
          { $tmpname[$count] = $fldnm[$i];
            $tmpsize[$count] = $width[$i];
            $count++;
          }
      }

    @fldnm = @tmpname;
    @width = @tmpsize;
    $count;

  }

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub runfiles

#
# Subroutine to run the loop through fits files and detached header files,
# and subdirectories as needed
#
#   Arguments:  $_[0] = directory (unopened)
#               $_[1] = $opt_a, check all files
#               $_[2] = $opt_r, recursion flag
#
  { local ($dir=$_[0], $aflag= $_[1], $rflag=$_[2]);
    local (@fits_file);                  # file names ending in hdr/fits/fit/fts
    local (@subdir);                     # subdirectory names
    local ($lf,$sd);                     # element counters for filename lists
    local ($file,$i);                    # loop/subscript variables
    local ($full_name);                  # full file name

    # Open the directory:

    opendir(DIR,$dir) || die "fitsidx: Could not open \"$dir\".\n";

    # Get a sorted list of files in the directory, and parse it into lists
    # of fits file names and subdirectories:

    $lf = $sd = 0;
    foreach $file (sort readdir(DIR))
      { if ($file =~ /^\./)
          { next; }
        else
          { $file = $dir . '/' . $file; }

        if (-d $file  &&  $rflag)     # Subdirectory check
          { $subdir[$sd] = $file;
            $sd++;
          }
        elsif ($aflag)                  # Processing all files
          { $fits_file[$lf] = $file;
            $lf++; 
          }
        else
          { $tfile = $file;
            $tfile =~ tr/a-z/A-Z/;
            if ($tfile =~ /\.HDR$/  ||  $tfile =~ /\.FITS$/  ||
                $tfile =~ /\.FIT$/  ||  $tfile =~ /\.FTS$/ )
              { $fits_file[$lf] = $file;
                $lf++;
              }
          }
      }

    #  Now process the FITS files:

    foreach $file (@fits_file)
      { open (FITS,$file) || die "fitsidx: Could not open \"$file\".\n";
        &process_fits($file);
        close(FITS);
      }

    # Close the current directory and descend through subdirectories:

    close(DIR);

    foreach $file (@subdir)
      { &runfiles($file,$aflag,$rflag); }
  }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
sub process_fits

#  Routine to read through the fits file and pull out the desired keywords.
#
#  Arguments: $_[0]  = the full name of the file (already opened on FITS)
#
  { local ($file=$_[0]);      # Full name of the file on <FITS>
    local ($line);            # input buffer
    local ($buffer);          # string buffer
    local ($found_first) = 0; # flag for first object pointer
    local (%val);             # Array of found value
    local ($more_header) = 0; # End-of-Header flag

    # Read the first line and make sure this is a FITS header:

    read(FITS,$line,80);
    if ($line !~ /^SIMPLE  =\s+T/)
      { return $file; }

    #  Assign the special values for ^PATH and ^FILE:

    $buffer = $file;
    $buffer =~ s/\/[a-zA-Z0-9_.]*$/\//;
    $val{'^PATH'} = $buffer;

    $buffer = $file;
    $buffer =~ s/^.*\///;
    $val{'^FILE'} = $buffer;

    $extension = "";
    while (read(FITS,$line,80))
      { $line =~ s/^\s+//;                    # Clip leading blanks

        # Start by checking for an END statement:

        if ($line =~ /^END /)
          { read(FITS,$line,80);

            # Found end of this header.  Look for an extension header vs
            # the beginning of data:

            while ($line =~ /^        /)  
              { read(FITS,$line,80); }

            if ($line =~ /^XTENSION/)
              { 
                # Get the type of extension:

                $line =~ s/\s+//g;
                ($a,$type) = split(/=/,$line);

                # Increment the extension count for this type of extension,
                # and set the $extension variable:

                $type_count{$type}++;
                $extension = $type . '[' . $type_count{$type} . ']';
              }

            # Otherwise we're done with this file:

            else
              { last; }
          }

        # Ignore all non-keyword lines:

        elsif ($line !~ /^.{8}=/)
          { next; }

        # Format keyword lines and split:

        else
          { # Remove comments.  This can be tricky, since the comment
            # delimiter (/) may appear in value strings:

            if ($line =~ /=\s+'/)
              { $line =~ s/^([^.]*'[^']').*$/$1/; }  # Drop everything after '
            else
              { $line =~ s/\/.*$//; }                # Drop everythign after /

            $line =~ s/\s*=\s+/=/;   # Collapse space around "="
            $line =~ s/\s+$//;       # Remove trailing blanks
            $line =~ s/['"]//g;      # Remove quotes

            ($keyword,$value) = split(/=/,$line);
            # Add the extension path, if any, to the keyword.  If this string
            # has a value in the %size array, save its corresponding value in 
            # the %val array:

            $keyword = join(".",$extension,$keyword) if ($extension);
            if ($size{$keyword})
              { $val{$keyword} = $value; }
          }
      }

    #
    #  End of label.  Print the index line for this label.  Values are printed
    #  in the order in which they were listed, default values first:
    #

    if ($opt_o)
      { $oldhandle = select(OUTPUT); }

    #  First, write the format string:

    for ($i=0; $i<$num_fields-1; $i++)
      { $bytes = $size{$name[$i]};
        if ($bytes < 0)
          { $format = sprintf ("%%%d.%ds ",$bytes,-$bytes); }
        else
          { $format = sprintf ("%%%d.%ds ",$bytes,$bytes); }
        printf $format,$val{$name[$i]};
      }

    $bytes = $size{$name[$num_fields-1]};
    if ($bytes < 0)
      { $format = sprintf ("%%%d.%ds ",$bytes,-$bytes); }
    else
      { $format = sprintf ("%%%d.%ds ",$bytes,$bytes); }
    printf $format,$val{$name[$num_fields-1]};
    printf "\n";
    if ($opt_o)
      { select($oldhandle); }

  }


#=============================================================================

# getopts.pl - a better getopt

# Usage:
#      do Getopts('a:bc');  
# -a takes arg. -b & -c not. Sets opt_* as a side effect.

sub Getopts {
    local($argumentative) = @_;
    local(@args,$_,$first,$rest);
    local($errs) = 0;
    local($[) = 0;

    @args = split( / */, $argumentative );
    while(@ARGV && ($_ = $ARGV[0]) =~ /^-(.)(.*)/) {
	($first,$rest) = ($1,$2);
	$pos = index($argumentative,$first);
	if($pos >= $[) {
	    if($args[$pos+1] eq ':') {
		shift(@ARGV);
		if($rest eq '') {
		    ++$errs unless @ARGV;
		    $rest = shift(@ARGV);
		}
		eval "\$opt_$first = \$rest;";
	    }
	    else {
		eval "\$opt_$first = 1";
		if($rest eq '') {
		    shift(@ARGV);
		}
		else {
		    $ARGV[0] = "-$rest";
		}
	    }
	}
	else {
	    print STDERR "Unknown option: $first\n";
	    ++$errs;
	    if($rest ne '') {
		$ARGV[0] = "-$rest";
	    }
	    else {
		shift(@ARGV);
	    }
	}
    }
    $errs == 0;
}

#
# User's notes:  if you get a rash of "Bad free() ..." messages, check
# the keyword index file used for proper justification.  Deleting a 
# "-" sign may solve the problem.

