#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

/* Program to fix a problem with non-printing characters in FITS headers, and
   adjust data-segment padding.

   These problems are specific to FITS files coming out of Flagstaff, AZ, from
   the UMD Hale-Bopp observing run.  Some of the FITS headers had newline 
   characters in the COMMENT fields.  This program searches the header for
   all non-printing characters and replaces them with blanks.  It double-checks
   byte counts in the header on exit.

   It also collected the NAXISn and BITPIX values and calculates the size of
   the data array, then ensures that the output data are correctly padded to 
   an integral number of FITS blocks.  It pads, when necessary, with nulls.

   This program assumes the input file is a simple, standard FITS image with
   only the primary data unit present.

   Format:

     % fixaz [ input file  [ output file ] ]

   I/O is through standard input/output if file names are omitted.

   03 April 1997, acr.
*/

main(argc,argv)
  int   argc;
  char *argv[];
  
{ FILE      *ifp, *ofp;     /* file pointers       */
  char       block[2880];   /* data block          */
  char       line[81];      /* header line buffer  */
  char      *p;             /* conversion pointer  */
  int        i,j,k;         /* loop/subscript      */
  long int   bitpix;        /* BITPIX value        */
  long int   databytes;     /* total data size     */
  long int   datablocks;
  long int   bcount;        /* block count         */
  long int   ival;          /* conversion value    */
  int        not_done;      /* loop flag           */

  /* Start by reading any parameters and opening files: */

  ifp = stdin;
  ofp = stdout;
  bcount = 0;

  if (argc > 3  ||  (argc == 2  &&  strcmp(argv[1],"?") == 0) )
    { fprintf (stderr,"Usage: %% fixaz [ input_file [ output_file ] ]\n");
      exit (0);
    }

  if (argc > 1)
    { if ((ifp = fopen(argv[1],"rb")) == NULL)
        { fprintf (stderr,"fixaz: Unable to open %s for input (%s).\n",
                           argv[1],strerror(errno));
          exit(10);
        }
    }

  if (argc > 2)
    { if ((ofp = fopen(argv[2],"wb")) == NULL)
        { fprintf (stderr,"fixaz: Unable to open %s for output (%s).\n",
                           argv[2],strerror(errno));
          exit(11);
        }
    }

  /*======================================================================*/

  /* Header processing */

  /* Read the first header block: */

  if (fread((void *)block,2880,1,ifp) != 1)
    { fprintf (stderr, "fixaz: Error reading from first header block, %s.\n",
                        strerror(errno));
      exit(100);
    }

  /* We assumed proper FITS format, in the the first lines of the header are
     SIMPLE, BITPIX, NAXIS, NAXIS1, etc., and collect the values needed:
  */

  /* Start with BITPIX: */

  p = block + 80;                             /* point to second line */
  if (strncmp(p,"BITPIX",6) != 0)
    { fprintf (stderr,"fixaz: Non-standard FITS - couldn't find BITPIX.\n");
      exit (110);
    }
  p += 10;                                    /* advance pointer past the '=' */
  bitpix = atol(p);
  bitpix = (bitpix < 0)? -bitpix : bitpix;    /* absolute value */
  if (bitpix < 8  ||  bitpix > 32  ||  bitpix % 8)
    { fprintf (stderr,"fixaz: Invalid BITPIX value (%d).\n",bitpix); 
      exit (120);
    }

  /* Determine the number of NAXIS values: */

  p = block + 160;
  if (strncmp(p,"NAXIS",5) != 0)
    { fprintf (stderr,"fixaz: Non-standard FITS - couldn't find NAXIS.\n");
      exit (130);
    }
  p += 10;
  ival = atol(p);
  if (ival < 1)
    { fprintf (stderr,"fixaz: Invalid NAXIS value (%d).\n",ival);
      exit(140);
    }

  /* Gather up the NAXIS values: */

  j = ival;
  databytes = bitpix/8;
  bcount = 1;

  for (i=0; i<j; i++)
    { p += 80;           /* move p to just past '=' of next line */
      ival = atol(p);
      if (ival < 1)
        { fprintf (stderr, "fixaz: Invalid NAXIS%d value (%d).\n",i+1,ival);
          exit(150);
        }
      databytes = databytes * ival;
    }

  /* OK, that's it for special processing.  Now start the loop that will
     dig through the header block, replacing non-printing characters with
     characters with blanks, and look for the END statement to end the
     loop:
  */

  not_done = 1;
  while (not_done)
    { for (i=0; i<2880; i++)
        { block[i] = (isprint(block[i]))? block[i] : ' '; }

      /* Check from the end of the block for the END statement and reset
         the flag if found:
      */

      p = block+2800;         /* point to beginning of last line */
      for (i=0; i<80; i++)
        { if (strncmp(p,"END     ",8) == 0)
            { not_done = 0;
              break;
            }
          p -= 80;
        }

      /* Write this header block: */

      if (fwrite((void *)block,2880,1,ofp) != 1)
        { fprintf (stderr,"fixaz: Error writing (header) block %d (%s).\n",
                           bcount, strerror(errno));
          exit(200);
        }

      /* Read next header block: */

      if (not_done)
        { if (fread((void *)block,2880,1,ifp) != 1)
            { fprintf (stderr,"fixaz: Error reading (header) block %d (%s).\n",
                               bcount+1, strerror(errno));
              exit(210);
            }
           bcount++;
        }
    }

  /*========================================================================*/

  /* Data Processing */

  /* Determine the number of whole blocks we expect to read: */

  datablocks = databytes / 2880;    /* Note that fractions were truncated */

  /* Whole data blocks are just transferred from the input file to the 
     output file:
  */

  for (i=0; i<datablocks; i++)
    { if (fread((void *)block,2880,1,ifp) != 1)
        { fprintf (stderr,"fixaz: Error reading (data) block %d (%s).\n",
                           bcount+1, strerror(errno));
          exit(300);
        }
      bcount++;

      if (fwrite((void *)block,2880,1,ofp) != 1)
        { fprintf (stderr,"fixaz: Error writing (data) block %d (%s).\n",
                           bcount, strerror(errno));
          exit(310);
        }
    }

  /* The last block may be properly padded or may end where the data do.  
     So, we'll just read the left-over bytes (if any) and pad with nulls
     ourselves:
  */

  ival = databytes - (datablocks * 2880);
  if (ival > 0)
    { if (fread((void *)block,ival,1,ifp) != 1)
        { fprintf (stderr,"fixaz: Error reading last (data) block, %d (%s).\n",
                           bcount+1, strerror(errno));
          exit(320);
        }
      bcount++;

      for (i=ival; i<2880; i++)
        { block[i] = 0; }

      if (fwrite((void *)block,2880,1,ofp) != 1)
        { fprintf (stderr,"fixaz: Error writing last (data) block, %d (%s).\n",
                           bcount, strerror(errno));
          exit(330);
        }
    }

  /* And we're done.  Close the files and exit: */

  if (ifp != stdin)
    { fclose(ifp); }
  if (ofp != stdout)
    { fclose(ofp); }

  exit (0);
}
