/*
  FLT2INT

  Program to convert a floating-point, conformant FITS file to an integer
  FITS file. The FITS header is updated appropriately, making use of the 
  BSCALE and BZERO keywords as needed.

  Format:

     % flt2int  <input_file> [ <output_file> ] 


  The input file must be a conformant FITS file. FLT2INT will convert the
  primary data unit and any IMAGE extensions found. The output is directed
  to stdout unless an output file name is provided. The output file will
  use 32 bit signed integers.

  22 Oct 2001, acr.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <values.h>

/*--------------------------------------------------------------------------*/
/* MACROS                                                                   */
/*--------------------------------------------------------------------------*/

#define  TRUE       1
#define  FALSE      0
#define  UNDEFINED -1


/*--------------------------------------------------------------------------*/
/* DATA STRUCTURES                                                          */
/*--------------------------------------------------------------------------*/

/* FITS data unit information: */

typedef struct funit
  { int           header_blocks;     /* Number pf header blocks */
    int           data_blocks;       /* Number of data blocks */
    int           bitpix;            /* BITPIX value */
    int           bsflag,bzflag;     /* TRUE if BSCALE and BZERO are present */
    double        max,min;           /* extrema */
    double        bscale,bzero;      /* existing BSCALE and BZERO */
    double        scale, zero;       /* new BSCALE and BZERO */
    char          xtension[15];      /* XTENSION keyword value */
    struct funit *next;
  } FITS_unit;


/*--------------------------------------------------------------------------*/
/* FUNCTIONS:                                                               */
/*--------------------------------------------------------------------------*/

FITS_unit *scan_fits_file(FILE *fits_file);
void       find_scale(FITS_unit *descriptor);
void       copy_asis(FITS_unit *descriptor, FILE *input, FILE *output);
void       convert_r4_data(FITS_unit *descriptor, FILE *input, FILE *output);
void       convert_r8_data(FITS_unit *descriptor, FILE *input, FILE *output);
int        end_of_header(char *header_block);
void       check_scale_zero(char *header_block, FITS_unit *descriptor);
void       find_fltext(FILE *infile, FITS_unit *descriptor);
void       find_dblext(FILE *infile, FITS_unit *descriptor);
void       write_header(FITS_unit *descriptor, FILE *input, FILE *output);


/*==========================================================================*/

main (int argc, char *argv[])

{ /* Arguments and options: */

  char   infile[500];    /* input file name */
  char   outfile[500];   /* output file name */

  FILE      *ifp,*ofp;   /* I/O pointers */
  FITS_unit *header;     /* pointer to top of linked list */

  char       buff[2880]; /* FITS block buffer */
  char       line[80];   /* Single header line buffer */

  FITS_unit *udsc;       /* FITS unit descriptor pointer */

  int        i,j,k;      /* loop/subscript */
  double     val;        /* calculation holders */
  long int   ival;

  /* We'll begin by checking the argument list and setting flags as needed: */

  if (argc < 2  ||  argc > 3)
    { fprintf (stderr,
               "Usage: flt2int <input_file> [ <output_file> ]\n");
      exit(0);
    }

  strcpy(infile,argv[1]);
  if ((ifp = fopen(infile,"rb")) == NULL)
    { fprintf (stderr,"Unable to open '%s' for reading (%s).\n",
               infile, strerror(errno));
      exit(100);
    }

  if (argc > 2)
    { strcpy(outfile, argv[2]); 
      if ((ofp=fopen(outfile,"wb")) == NULL)
        { fprintf (stderr,"Unable to open '%s' for writing (%s)\n",
                   outfile, strerror(errno));
          exit(110);
        }
    }
  else
    { ofp = stdout; }

  /* All right, we're good to go. In order to maximize processing and do this
     in the minimum number of passes (2), we're going to read through and 
     analyze the entire data file first, collecting information on header 
     types, extrema, and any existing BSCALE and BZERO values. This information
     is saved into the 'hdr' linked list:
  */

  header = scan_fits_file(ifp);


  /* And we'll call a routine to calculate the new BSCALE and BZERO for 
     each applicable data unit:
  */

  udsc = header;

  while (udsc != NULL)
    { if (udsc->bitpix < 0)    /* i.e., floating point numbers */
        { find_scale(udsc); }

      udsc = udsc->next;
    }

/* Debug printing...

udsc = header;
i = 1;
while (udsc)
  { fprintf (stderr,"\nBlock %d (%s):\n",i,udsc->xtension);
    fprintf (stderr,"  header blocks: %d\n    data blocks: %d\n",
            udsc->header_blocks,udsc->data_blocks);
    fprintf (stderr,"         bitpix: %d\n",udsc->bitpix);
    if (udsc->bitpix < 0)
      { fprintf (stderr,"            max: %e\n",udsc->max);
        fprintf (stderr,"            min: %e\n",udsc->min);
        if (udsc->bsflag) fprintf (stderr,"         bscale: %f\n",udsc->bscale);
        if (udsc->bzflag) fprintf (stderr,"          bzero: %f\n",udsc->bzero);
        fprintf (stderr,"      New scale: %e\n",udsc->scale);
        fprintf (stderr,"       New zero: %e\n",udsc->zero);
        val = udsc->max * udsc->bscale + udsc->bzero;
        ival = (val - udsc->zero)/udsc->scale;
        fprintf (stderr,"        New max: %d\n",ival);
        val = udsc->min * udsc->bscale + udsc->bzero;
        ival = (val - udsc->zero)/udsc->scale;
        fprintf (stderr,"        New min: %d\n",ival);

      }

    udsc = udsc->next;
  }
...Debug printing */

  /* Now we're ready to write the output file. We'll close and re-open the
     input file to reset it for reading:
  */

  fclose(ifp);
  ifp = fopen(infile,"rb");

  udsc = header;
  while (udsc != NULL)
    { if (udsc->bitpix >= 0)
        { copy_asis(udsc,ifp,ofp); }
      else if (udsc->bitpix == -32)
        { convert_r4_data(udsc,ifp,ofp); }
      else
        { convert_r8_data(udsc,ifp,ofp); }
      udsc = udsc->next;
    }

  /* All done - close up and go home: */

  fclose(ifp);
  fclose(ofp);

  exit(0);
}

/*==========================================================================*/
/*==========================================================================*/
/* SUBROUTINES                                                              */
/*==========================================================================*/
/*==========================================================================*/

FITS_unit *scan_fits_file(FILE *ifp)

  /* This routine reads through the entire FITS file, gathering information
     about the data units into a linked list. A pointer to the header of 
     the list is returned.
  */

{ FITS_unit *top;        /* Top-of-list pointer */
  FITS_unit *new,*last;

  char       buff[2880];       /* input block buffer */
  char       line[80];         /* input line buffer  */
  int        bytes;            /* number of bytes read */
  int        last_block;       /* TRUE if this header has an END statement */
  int        naxis,naxisn[7];  /* NAXIS and NAXISn values */
  long int   size;

  int        i,j,k;            /* loop/subscript */
  char      *p,*q;             /* utility poiners */

  top = NULL;

  /* There must be at least one data unit, so we'll read in the first
     block, do some sanity checking and start allocating list space:
  */

  bytes = fread((void *)buff,1,2880,ifp);
  if (bytes != 2880  ||  strncmp(buff,"SIMPLE",6) != 0)
    { fprintf (stderr,"ERROR: Input file is not a FITS file.\n");
      fclose(ifp);
      exit(200);
    }

  /* Now we'll start a loop that will run once for each data unit: */

  while (bytes > 0)
    { /* Allocate space for the new descriptor and initialize: */

      new = (FITS_unit *)malloc(sizeof(FITS_unit));
      new->header_blocks = 1;
      new->data_blocks   = 0;
      new->max           = MINDOUBLE;
      new->min           = MAXDOUBLE;
      new->bsflag        = FALSE;
      new->bzflag        = FALSE;
      new->bscale        = 1.0;
      new->bzero         = 0.0;
      new->next          = NULL;

      /* Set the extension type: */

      if (top == NULL)
        { strcpy(new->xtension,"IMAGE"); }
      else
        { strncpy(line,buff,80);
          p = line;
          while (p[0] != '\'') p++;
          p++;
          i = 0;
          while (p[i] != '\'') i++;
          p[i] = '\0';
          strcpy(new->xtension,p);
        }

      /* Determine BITPIX: */

      p = buff+80+10;      /* We'll use 'p' to step through values */

      new->bitpix = atoi(p);

      /* Get the number of axes and their sizes and calculate data blocks: */

      p += 80;
      naxis = atoi(p);
      size = abs((new->bitpix)/8);

      for (i=0; i<naxis; i++)
        { p+= 80;
          naxisn[i] = atoi(p);
          size = size * naxisn[i];
        }
      new->data_blocks = (size % 2800)? (size/2880) + 1 : size/2880;

      /* Check for end of header: */

      last_block = end_of_header(buff);


      /* All right. If this is not real data, then we read past the header
         and data to the next header:
      */

      if (new->bitpix > 0)
        { while (!last_block)
            { fread((void *)buff,2880,1,ifp);
              new->header_blocks++;
              last_block = end_of_header(buff);
            }
          for (i=0; i<new->data_blocks; i++)
            { fread((void *)buff,2880,1,ifp); }
        }

      /* But if it is real data, we need to keep an eye out for BSCALE
         and BZERO as we read past the header:
      */

      else
        { check_scale_zero(buff,new); 
          while (!last_block)      
            { fread((void *)buff,2880,1,ifp);
              new->header_blocks++;
              last_block = end_of_header(buff);
              check_scale_zero(buff,new);
            }

          /* And we'll call a routine to collect max/min for the data: */

          if (new->bitpix == -32)
            { find_fltext(ifp,new); }
          else
            { find_dblext(ifp,new); }
        }

      /* Add the new unit descriptor to the linked list: */

      if (top == NULL)
        { top = new;
          last = new;
        }
      else
        { last->next = new;
          last = new;
        }

      /* Now we attempt to read the next block and loop: */

      bytes = fread((void *)buff,1,2880,ifp);
    }

  /* Return the pointer to the head of the list and we're done: */

  return top;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int end_of_header(char *hdr)

/* Routine to check the header block passed in for an "END" statement. If an
   "END" is found, the line number is returned.
*/

{ int     i;     /* loop/subscript */

  /* We're going to step through the block backwards, since the END statement
     is slightly more likely to occur at the end of the block:
  */

  for (i = 2800; i>= 0; i = i-80)
    { if (strncmp(hdr+i,"END ",4) == 0)
        { return (i/80)+1; }
    }

  /* If we mad it this far, we have no "END": */

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void check_scale_zero(char *hdr, FITS_unit *dsc)

/* Routine to check the input header block for BSCALE and BZERO keywords,
   setting the appropriate values in the unit descriptor when found.
*/

{ int     i;  /* loop/subscript */
  char   *p;  /* convenience pointer */

  /* We'll check each line for the desired keywords: */

  for (i=0; i<2880; i=i+80)
    { p = hdr+i;
      if (strncmp(p,"BSCALE ",7)==0)
        { dsc->bscale = atof(p+10); 
          dsc->bsflag = TRUE;
        }
      else if (strncmp(p,"BZERO ",6)==0)
        { dsc->bzero  = atof(p+10); 
          dsc->bzflag = TRUE;
        }
    }

  /* All done: */

  return;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void find_fltext(FILE *ifp, FITS_unit *dsc)

/* Routine to find the extrema in data blocks of type 'float' (BITPIX = -32).*/

{ float   data[720];   /* Block buffer */
  int     i,j,k;       /* loop/subscript */

  /* Loop through the expected data blocks: */

  for (i=0; i<dsc->data_blocks; i++)
    { fread((void *)data,2880,1,ifp);

      /* Check each value for extremes: */

      for (j=0; j<720; j++)
        { dsc->max = (data[j] > dsc->max)? (double)data[j] : dsc->max;
          dsc->min = (data[j] < dsc->min)? (double)data[j] : dsc->min;
        }
    }

  /* Done. */

  return;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void find_dblext(FILE *ifp, FITS_unit *dsc)

/* Routine to find the extrema in data blocks of type 'double' (BITPIX = -64).*/

{ double  data[360];   /* Block buffer */
  int     i,j,k;       /* loop/subscript */

  /* Loop through the expected data blocks: */

  for (i=0; i<dsc->data_blocks; i++)
    { fread((void *)data,2880,1,ifp);

      /* Check each value for extremes: */

      for (j=0; j<360; j++)
        { dsc->max = (data[j] > dsc->max)? data[j] : dsc->max;
          dsc->min = (data[j] < dsc->min)? data[j] : dsc->min;
        }
    }

  /* Done. */

  return;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void       find_scale(FITS_unit *dsc)

/* Routine to determine the new BSCALE and BZERO values for the converted
   real data blocks. The output will be 32-bit signed integers, so this is
   merely a process of mapping the extremes of the physical values to the
   extremes of the 'long int' range (+/- MAXLONG).
*/

{ double  max,min;        /* Modified extrema */

  /* First, if there is a BSCALE or BZERO we need to determine the physical
     extrema:
  */

  if (dsc->bsflag)
    { max = dsc->max * dsc->bscale;
      min = dsc->min * dsc->bscale;
    }
  else
    { max = dsc->max;
      min = dsc->min;
    }

  if (dsc->bzflag)
    { max = max + dsc->bzero;
      min = min + dsc->bzero;
    }

  /* Now we calculate the new scale and offset values: */

  dsc->scale = (max - min)/(2.0 * MAXLONG);
  dsc->zero  = (max + min)/2.0;


  return; 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void       copy_asis(FITS_unit *dsc, FILE *ifp, FILE *ofp)

/* Routine to copy a header and data unit without modification. */

{ char    buff[2880];   /* Block buffer */
  int     i,j,k;        /* loop/subscript */

  /* First we copy the header blocks, then the data blocks: */

  for (i=0; i<dsc->header_blocks; i++)
    { fread((void *)buff,2880,1,ifp);
      fwrite((void *)buff,2880,1,ofp);
    }

  for (i=0; i<dsc->data_blocks; i++)
    { fread((void *)buff,2880,1,ifp);
      fwrite((void *)buff,2880,1,ofp);
    }

  /* Done. */

  return;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void       convert_r4_data(FITS_unit *dsc, FILE *ifp, FILE *ofp)

/* Routine to convert 'float' data as it is written out. */

{ float    inbuff[720];  /* float input buffer */
  long int i4buff[720];  /* integer output buffer */
  int      i,j,k;        /* loop/subscript */
  double   val;          /* conversion buffer */
  double   bscale,bzero; /* convenience variables */
  double   scale,zero;

  /* First we'll write the header. Since we need to modify the header for
     the new BSCALE/BZERO values, we'll call a subroutine:
  */

  write_header(dsc,ifp,ofp);

  /* Now we'll loop through the real data blocks. For each value, we'll
     apply any existing BSCALE and BZERO, the convert the number using
     the new BSCALE and BZERO and forcing it to an integer:
  */

  bscale = dsc->bscale;
  bzero  = dsc->bzero;
  scale  = dsc->scale;
  zero   = dsc->zero;

  for (i=0; i<dsc->data_blocks; i++)
    { fread((void *)inbuff,2880,1,ifp);

      for (j=0; j<720; j++)
        { val = inbuff[j] * bscale + bzero;
          i4buff[j] = (long int)((val - zero)/scale);
        }

      fwrite((void *)i4buff,2880,1,ofp);
    }

  /* Done: */

  return; 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void       convert_r8_data(FITS_unit *dsc, FILE *ifp, FILE *ofp)

/* Routine to convert 'float' data as it is written out. */

{ float    inbuff[360];  /* double input buffer */
  long int i4buff[360];  /* integer output buffer */
  int      i,j,k;        /* loop/subscript */
  double   val;          /* conversion buffer */
  double   bscale,bzero; /* convenience variables */
  double   scale,zero;

  /* First we'll write the header. Since we need to modify the header for
     the new BSCALE/BZERO values, we'll call a subroutine:
  */

  write_header(dsc,ifp,ofp);

  /* Now we'll loop through the real data blocks. For each value, we'll
     apply any existing BSCALE and BZERO, the convert the number using
     the new BSCALE and BZERO and forcing it to an integer:
  */

  bscale = dsc->bscale;
  bzero  = dsc->bzero;
  scale  = dsc->scale;
  zero   = dsc->zero;

  for (i=0; i<dsc->data_blocks; i++)
    { fread((void *)inbuff,2880,1,ifp);

      for (j=0; j<360; j++)
        { val = inbuff[j] * bscale + bzero;
          i4buff[j] = (long int)((val - zero)/scale);
        }

      fwrite((void *)i4buff,1440,1,ofp);
    }

  /* Since we're writing half-blocks here, we need to check to see if the
     number of input data blocks was even. If not, we'll need to pad out 
     the last output FITS data block with nulls:
  */

  if (dsc->data_blocks % 2)
    { memset((void *)i4buff,1440,'\0');
      fwrite((void *)i4buff,1440,1,ofp);
    }

  /* Done: */

  return; 
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void write_header(FITS_unit *dsc, FILE *ifp, FILE *ofp)

/* Routine to modify the output FITS header using the new BITPIX, BSCALE
   and BZERO values.
*/

{ char    inbuff[2880];                    /* Input buffer */
  char    outbuff[2880];                   /* output buffer */
  char    bscale[80],bzero[80],bitpix[80]; /* record buffers */
  char    line[80]; 
  int     bytes;                           /* byte count */
  int     eoh;                             /* end-of-header indicator */
  int     inlines;                         /* lines to be inserted */

  int     i,j,k;                           /* loop/subscript */
  char   *p;                               /* utility pointer */

  /* We'll begin by preparing the lines we're going to insert: */

  sprintf (bitpix,"BITPIX  = %20d",32);
  sprintf (bscale,"BSCALE  = %20.10e",dsc->scale);
  sprintf (bzero, "BZERO   = %20.10e",dsc->zero);

  memset(bitpix+30,' ',50);
  memset(bscale+30,' ',50);
  memset(bzero+30, ' ',50);

  /* We'll loop through the header blocks: */

  for (i=0; i<dsc->header_blocks; i++)
    { fread((void *)inbuff,2880,1,ifp);

      /* Replace the BITPIX line in the first block: */

     if (i==0) memcpy(inbuff+80,bitpix,80); 

      /* If there was an original BSCALE or BZERO line, we check this block 
         for it and replace it:
      */

      if (dsc->bscale  ||  dsc->bzero)
        { for (j=0; j<2880; j=j+80)
            { p = inbuff+j;
              if (dsc->bsflag  &&  strncmp(p,"BSCALE ",7)==0)
                { memcpy(p,bscale,80); }
              if (dsc->bzflag  &&  strncmp(p,"BZERO  ",7)==0)
               { memcpy(p,bzero,80); }
            }
        }

      /* If there wasn't an original BSCALE or BZERO, we will insert the lines
         before the final END statement, because this is easier. So, unless
         this is the last block, we write it out and continue the loop:
      */

      if (i < dsc->header_blocks-1)
        { fwrite((void *)inbuff,2880,1,ofp); }
    }

  /* Now, we've got the last block loaded up and ready to go. But if we need
     to insert lines, they go in here. First, determine how many lines we 
     need to insert:
  */

  inlines = 0;
  if (!dsc->bsflag) inlines++;
  if (!dsc->bzflag) inlines++;

  /* If there are no lines to insert, we write the block and we're done:

  if (inlines == 0)
    { fwrite((void *)inbuff,2880,1,ofp);
      return;
    }

  /* Now comes the tricky part. First we find out where the 'END' statement
     is:
  */

  eoh = end_of_header(inbuff);

  /* Now we see if there's enough room to insert our lines without overflowing
     the header:
  */

  if (36-eoh >= inlines)
    { /* There is!  So we'll overwrite the 'END' line and re-write it after
         the lines we add (remember FITS headers are padded with blanks):
      */

      i = (eoh-1);
      p = inbuff + (i*80);
      if (!dsc->bsflag)
        { memcpy(p,bscale,80);
          p = p+80;
        }
      if (!dsc->bzflag)
        { memcpy(p,bzero,80);
          p = p+80;
        }
      strncpy(p,"END ",4);

      /* And now we write the block: */

      fwrite((void *)inbuff,2880,1,ofp);
    }

  else
    { /* There isn't. To make things easier we will overwrite the 'END'
         statement with a blank line, then just start a new block to 
         hold the new lines and the displaced 'END':
      */

      i = (eoh-1);
      p = inbuff + (i*80);
      memcpy(line,p,80);
      memset(p,' ',80);

      fwrite((void *)inbuff,2880,1,ofp);

      /* New block: */

      i = 0;
      if (!dsc->bsflag)
        { memcpy(inbuff,bscale,80);
          i++;
        }
      if (!dsc->bzflag)
        { memcpy(inbuff+(i*80),bzero,80);
          i++;
        }
      memcpy(inbuff+(i*80),line,80);
      i++;

      /* Pad the remainder of the block with blanks: */

      while (i<36)
        { memset(inbuff+(i*80),' ',80);
          i++;
        }

      /* Now write the last block: */

      fwrite((void *)inbuff,2880,1,ofp);
    }

  /* And we're done: */

  return;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

