/*
   Program: make_fitstab

   Routine to complete the creation of a FITS table after running the 
   pds2fits routine to produce the required header file.  This routine does
   the following:

      1. Removes carriage-control from the PDS data file.
      2. Subtracts 2 from the NAXIS1 value in the FITS TABLE extension
         header to account for the new record size
      3. Writes the modified header and data into a single output file
         without carriage control
      4. Pads the output file with blanks to a multiple of 2880 bytes.

  Format:

     % make_fits <header file> <data file> [<output file>]

  14 Jul 1995, acr: creation

*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#define BLOCKSIZE 2880

/*=========================================================================*/

main(argc,argv)
  int    argc;
  char  *argv[];

{ FILE   *hfp, *dfp, *ofp;  /* Header, data and output file pointers */

  char   hdr[36][80];       /* header block holding array */
  char   iblock[BLOCKSIZE]; /* data input array */
  char   oblock[BLOCKSIZE]; /* data output array */

  int    i,j,k;             /* loop/subscript */
  int    bytes;             /* i/o byte count */
  int    done;              /* logical flag */

  /*---------------------------------------------------------------------*/
  /* Get the input arguments and open the files:                         */

  if (argc < 3  ||  argc > 4)
    { fprintf (stderr,"Usage: make_fitstab <header> <data> [<output>]\n");
      exit(0);
    }

  if ((hfp=fopen(argv[1],"rb")) == NULL)
    { fprintf (stderr,"make_fitstab: Unable to open %s for input (%s).\n",
                       argv[1], strerror(errno));
      exit(10);
    }

  if ((dfp=fopen(argv[2],"rb")) == NULL)
    { fprintf (stderr,"make_fitstab: Unable to open %s for input (%s).\n",
                       argv[2], strerror(errno));
      exit(11);
    }

  if (argc == 4)
    { if ((ofp=fopen(argv[3],"wb")) == NULL)
        { fprintf (stderr,
                   "make_fitstab: Unable to open %s for writing (%s).\n",
                   argv[3], strerror(errno));
          exit(12);
        }
    }
  else
      ofp = stdout;

  /*---------------------------------------------------------------------*/
  /* Process header blocks:                                              */

  while (fread((void *)hdr,BLOCKSIZE,1,hfp) == 1)
    { if (strncmp(hdr[0],"XTENSION",8)==0)
        
        /* Found the extension header. */

        { if (strncmp(hdr[0]+11,"TABLE",5) != 0)
            { fprintf (stderr, "make_fitstab:  Wait a minute - ");
              fprintf (stderr, "this XTENSION isn't a TABLE!\n");
              fprintf (stderr, "I can't handle complex FITS files...\n");
              exit(100);
            }

          /* Now we know it's a TABLE, so the fourth line must contain
             NAXIS1:
          */

          if (strncmp(hdr[3],"NAXIS1",6) != 0)
            { fprintf (stderr, "make_fitstab: NAXIS1 wasn't found where ");
              fprintf (stderr, "it should have been in the XTENSION header.\n");
              exit(101);
            }

          /* Read the NAXIS1 value in columns 11-30, subtract 2, and write it
             back:
          */

          i = atoi(hdr[3]+10) - 2;
          j = hdr[3][30];   /* save this value - it's about to get squashed */
          sprintf (hdr[3]+10,"%20d",i);
          hdr[3][30] = j;
        }

      /* Either way, write the block out: */

      if (fwrite((void *)hdr,BLOCKSIZE,1,ofp) != 1)
        { fprintf (stderr, "make_fitstab: Error on output (%s).\n",
                            strerror(errno));
          exit(102);
        }
    }

  /* Close the header file:

  fclose(hfp);

  /*---------------------------------------------------------------------*/
  /* Process data blocks:                                                */

  done = 0;
  bytes=fread((void *)iblock, 1, BLOCKSIZE,dfp);
  i=0;
  j=0;

  while (!done)
    { /* if this is a carriage-control character, skip it; otherwise
         transfer it to the output block:
      */

      if (iblock[i] == '\r'  ||  iblock[i] == '\n')
          i++;
      else
          oblock[j++] = iblock[i++];

      if (i == bytes)                        /* done with this block     */
        { if (bytes == BLOCKSIZE)            /* it was a whole block     */
            { bytes = fread((void *)iblock, 1, BLOCKSIZE,dfp); /* next block */
              i = 0;
              if (bytes == 0) done = 1;      /* no more data             */
            }
          else                               /* it wasn't a whole block  */
              done = 1;
        }
      if (j == BLOCKSIZE)                    /* filled this block        */
        { if (fwrite((void *)oblock,BLOCKSIZE,1,ofp) != 1)
            { fprintf (stderr, "make_fitstab: Error on output (%s).\n",
                                strerror(errno));
              exit(201);
            }
          j = 0;
        }
    }

  /* Check for a partial block still in 'oblock', pad with blanks and
     output:
  */

  if (j != 0)
    { for (i=j; i<BLOCKSIZE; ++i) oblock[i] = ' ';
      if (fwrite((void *)oblock,BLOCKSIZE,1,ofp) != 1)
        { fprintf (stderr, "make_fitstab: Error on output (%s).\n",
                            strerror(errno));
          exit(202);
        }
    }

  /*---------------------------------------------------------------------*/
  /* Done:                                                               */

  fclose(dfp);
  fclose(ofp);
  exit(0);
}

