/*
  primfits - return primary header and data from a FITS file.

  This routine takes a FITS file which may have extension headers and data
  and returns just the primary header and data.  It assumes properly (i.e.
  according to the standard) formatted FITS in the primary header and data.

  Format:   % primfits [ input [ output ] ]

  28 September 1995, acr.

*/

#include <stdio.h>
#include <string.h>
#include <errno.h>

#define FITSBLOCK  2880

main(argc,argv)
  int   argc;
  char *argv[];

{ FILE  *ifp,*ofp;          /* input/output file pointers */
  char  block[FITSBLOCK];   /* one FITS block */
  char  hdrline[81];        /* one line from the header */
  int   inheader;           /* logical flag */
  int   bcount;             /* block count */

  int   i,j,k;              /* loop/subscript */

  /* First, check for input arguments and open the appropriate files: */

  if (argc > 3)
    { fprintf (stderr,"Usage: primfits [ input [ output ] ]\n.");
      exit(0);
    }

  if (argc > 1)    /* i.e., there is an input file given */
    { if ((ifp = fopen(argv[1],"r")) == NULL)
        { fprintf (stderr, "Could not open %s for input (%s).\n",
                           argv[1], strerror(errno));
          exit(10);
        }
    }
  else
      ifp = stdin;

  if (argc > 2)   /* i.e., there is an output file given */
    { if ((ofp = fopen(argv[2],"w")) == NULL)
        { fprintf (stderr, "Could not open %s for writing (%s).\n",
                            argv[2], strerror(errno));
          exit(11);
        }
    }
  else
      ofp = stdout;

  /*---------------------------------------------------------------------*/

  /* This program uses a very simple algorithm to avoid math problems
     involved with pre-calculating the number of bytes in vary large files.
     It does assume, however, that the primary header and data conform to
     the FITS standard.

     So, the basic method is to start with the header blocks, checking
     each line in each block until the 'END' record is found.  Then, 
     the data blocks are transferred one at a time until either EOF, or
     and extension header is found.  The output file is closed at that
     point and the program stops.
  */

  inheader = 1;
  bcount   = 1;

  /* Header blocks: */

  while (inheader)
    { /* Read the next block: */

      if (fread((void *)block, 1, FITSBLOCK, ifp) != FITSBLOCK)
        { fprintf (stderr, "Read error on input block %d (header).\n",bcount);
          exit(20);
        }

      /* Search for END record: */

      for (i=0; i<36; ++i)
        { if (strncmp(block+(i*80),"END     ",8) == 0)
            { inheader = 0;
              break;
            }
        }

      /* Copy this block to the output file: */

      if (fwrite((void *)block, 1, FITSBLOCK, ofp) != FITSBLOCK)
        { fprintf (stderr, "Write error on output block %d (header).\n",bcount);
          exit(21);
        }

      bcount++;  /* next block */
    }


  /* Data blocks: */

  if (fread((void *)block, 1, FITSBLOCK, ifp) != FITSBLOCK)
    { fprintf (stderr, "Read error on input block %d (data),\n", bcount);
      exit(22);
    }

  while (!inheader  &&  !feof(ifp))
    { /* Check for beginning of extension header: */

      if (strncmp(block,"XTENSION",8) == 0)
        { inheader = 1; 
          break;         /* leave loop at next header */
        }

      /* Otherwise, write the block: */

      if (fwrite((void *)block, 1, FITSBLOCK, ofp) != FITSBLOCK)
        { fprintf (stderr, "Write error on output block %d (data).\n",
                            bcount);
          exit(23);
        }
      bcount++;

      /* Read the next block: */

      if (fread((void *)block, 1, FITSBLOCK, ifp) != FITSBLOCK  &&  !feof(ifp))
        { fprintf (stderr, "Read error on input block %d (data).\n", bcount);
          exit(24);
        } 
    }

  /* Display a warning in the case of EOF termination: */

  if (feof(ifp))
    { fprintf (stderr,"primfits: NOTE - this file did not contain any ");
      fprintf (stderr,"extension headers.\n");
    }

  /* Close the files and exit: */

  fclose(ifp);
  fclose(ofp);
  exit(0);
}
