/*
  sepfits: Routine to separate a FITS file with extension headers into 
           separate files.  The files generated from the extension headers
           will NOT have primary data headers attached (and thus are not
           true FITS files).

  Format:  % sepfits <input file>

  The output files will have the same filename as the input file, with an
  extension of ".pdu" for the primary data, and ".x##" for each extension,
  where ## is the sequential number of the extension header encountered.

  10 October 1995, acr.

*/

#include <stdio.h>
#include <string.h>
#include <errno.h>

#define FITSBLOCK  2880


main(argc,argv)
  int    argc;
  char  *argv[];

{ FILE *ifp, *ofp;      /* Input/output file pointers */
  char  filename[50];   /* filename holder */
  char  file[60];       /* full filename */
  size_t bytes;         /* string length */
  char  block[2880];    /* input block buffer */
  int   type,last_ext;  /* logical flags for block type*/

  int   i,j,k;          /* loop/subscript */


  /* First, check for an input file and open it: */

  if (argc != 2)
    { fprintf (stderr,"Usage: sepfits <input file>.\n");
      exit(0);
    }

  if ((ifp=fopen(argv[1],"r")) == NULL)
    { fprintf (stderr,"sepfits: Could not open %s for reading (%s).\n",
                      argv[1],strerror(errno));
      exit(10);
    }

  /* Break off and keep the file name of the input file: */

  bytes = strcspn(argv[1],".");
  strncpy(filename,argv[1],bytes);
  filename[bytes] = '\0';


  /* Start with the primary data unit header: */

  strcpy(file,filename);
  strcat(file,".pdu");
  if ((ofp=fopen(file,"w"))==NULL)
    { fprintf (stderr,"sepfits: Could not open %s for writing (%s).\n",
                       file,strerror(errno));
      exit(20);
    }

  
  /* Work through file: */

  /* ...primary header block... */

  type = readblock(block,ifp);
  while (type == 1)
    { fwrite ((void *) block, FITSBLOCK, 1, ofp);
      type = readblock(block,ifp);
    }

  /* ...primary data blocks (if any)... */

  while (type == 0)
    { fwrite ((void *) block, FITSBLOCK, 1, ofp);
      type = readblock(block,ifp);
    }

  last_ext = 1;

  /* ...extensions... */

  while (type >= 0)
    { /* Check for new extension: */

      if (type > 0  &&  type != last_ext)
        { fclose(ofp);
          sprintf (file,"%s.x%2.2d",filename,type-1);   /* open new file */
          if ((ofp=fopen(file,"w")) == NULL)
            { fprintf (stderr,"sepfits: Unable to open %s for writing (%s).\n",
                               file, strerror(errno));
              exit(30);
            }
          last_ext = type;
         }

      /* write the block and get the next one: */

      fwrite ((void *)block,FITSBLOCK,1,ofp);
      type = readblock(block,ifp);
    }

  /* Close the last file: */

  fclose(ofp);

  /* Done: */

  exit(0);

}


/*=========================================================================*/

int readblock(buffer,ifp)
  char  *buffer;        /* the input block buffer */
  FILE  *ifp;           /* input file pointer     */

  /* This routine reads through the blocks in a FITS file, keeping track of
     whether it is in the primary data header, data, or extension header
     blocks.  Status values return the type of block:

           -1  End of file (no block returned)
            0  Data
            1  Primary Data Header
           >1  Extension Header (# indicates the number of the extension)
  */

  { static  int  blocktype = -2;  /* initial value indicates new file */
    static  int  endheader = 0;   /* logical flag for end of header   */
    static  int  xcount = 1;      /* extension header counter         */

    int     i;                    /* loop/subscript */


    /* Read the next block into the buffer: */

    if (fread((void *)buffer, 1, FITSBLOCK, ifp) != FITSBLOCK)
      { /* Check for end of file: */

        if (feof(ifp)) 
            return -1;
        else
          { fprintf (stderr,"sepfits: Error reading from input file (%s).\n",
                             strerror(errno));
            exit(100);
          }
      }


    /* Determine the block type.  There are several cases:

         o  This is the first invocation.  Make sure the primary header
            begins with the "SIMPLE" statement.
         o  We've been reading data.  Check for the beginning of an 
            extension header.
         o  We're reading a header.  Look for the "END" statement.
         o  We've been reading one header, but reached the "END" statement 
            last time.  Check for the immediate beginning of another extension
            header (it is valid to have no data associated with a header).
            Otherwise, we're reading data.
    */

    if (blocktype == -2)
      { if (strncmp(buffer,"SIMPLE",6) == 0)
          { blocktype = 1;
            endheader = 0;
          }
        else
          { fprintf (stderr,"sepfits: Not a FITS file.\n");
            exit(101);
          }
      }

    else if (blocktype == 0)    /* Check for end of data */
      { if (strncmp(buffer,"XTENSION",8) == 0)
          { blocktype = ++xcount;
            endheader = 0;
          }
      }
           
    else if (blocktype > 0  &&  !endheader)
      { for (i=0; i<36; i++)
          { if (strncmp(buffer+(i*80),"END     ",8) == 0)
              { endheader = 1;
                break;
              }
          }
      }

    else  /* blocktype > 0  &&  endheader */
      { if (strncmp(buffer,"XTENSION",8) == 0)  /* another extension */
          { blocktype = ++xcount;
            endheader = 0;
          }
        else
          { blocktype = 0; }                    /* reading data now */
      }
  
    /* Return the block type and we're done: */

    return blocktype;
}
