#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#define PERM            0644       /* File creation permissions */
#define DEFAULT_DELIM      0       /* Default delimiter style, 0=Unix */

/*

  splitfits:  Routine to split up a FITS file into more managable pieces.

  This routine takes a FITS file as input and splits it into separate header
  and data files.  The user may indicate that record delimiters should be 
  added (Unix, PC or Mac style) to header, data or both.  Where to place 
  the delimiters is determined by the corresponding keyword in the FITS
  primary or extension header.  In addition, extensions are separated from
  the primary data and each other.  In files where there is no primary data,
  the primary header and first extension header are combined in a single
  output header file.

  Output file names are generated by the program, using the input file name
  and the following file extensions:

      .h##      Header
      .d##      Data
  
  where "##" is a sequential number starting with the first header unit with
  associated data.  In most cases, this will be the primary data unit, but in
  cases where there is no pimary data but there are extensions, the first
  set of files will contain the first extension data.


  Format:

      $ splitfits [-d delimiter] [-h] [-r] [-s] FITS_file

  Options:

     -d delimiter   Type of record delimiters to use ("unix","mac" or "pc");
                    default is "unix"

     -h             Insert record delimiters into header files every 80 bytes

     -r             Insert record delimiters into data files every NAXIS1 bytes

     -s             Silent.  Do not print status messages after successful
                    processing.


   Modification History:

   27 Jan 1997, acr: rewritten from scratch to fix a few bugs and add some
                     functionality.

*/

int get_naxis1(char *, int);
int end_block(char *);
int next_block(FILE *, char *, int *);
int write_block(FILE *, char *, int, char *, int);


main (int argc, char *argv[])

{ /* File parameters: */

  FILE   *ifp,*ofp;          /* File pointers */
  int     naxis;             /* number of axis in data */
  int     naxis1;            /* data line length */
  char    fitsname[250];     /* input FITS file name */
  char    filename[250];     /* filename buffer */
  int     blocknum;          /* input block count */
  int     no_primary_data;   /* flag */

  /* Input parameters: */

  char    block[2880];       /* input FITS block */
  int     delimiter;         /* 0=\n; 1=\r; 2=\r\n */
  int     noisy;             /* 1 unless -s option given */
  int     breakheader;       /* 80 if adding cc to header, else 0 */
  int     breakdata;         /* NAXIS1 if adding cc to data, else 0 */

  /* Output counters: */

  int     unitnum;           /* unit number for creating output files */
  int     fileblock;         /* Number of blocks output so far */

  /* Program variables: */

  int     i,j,k;             /* loop/subscript */
  char    hline[85];         /* header line */
  char   *dline;             /* data line pointer (must be allocated) */
  char   *sptr;              /* generic string pointer */
  size_t  nbytes;            /* I/O bytes */
  int     not_done;          /* loop flag */
  int     not_eof;           /* loop flag */
  char    write_perm[10];    /* file permissions string */


  /* Initializations: */

  delimiter   = 0;
  noisy       = 1;
  breakheader = 0;
  breakdata   = 0;
  unitnum     = 0;
  blocknum    = 0;

  /*========================================================================*/

  /* Start by reading through the arguments and parameters: */

  if (argc < 2  || argc > 7)
    { fprintf (stderr,
              "Usage: splitfits [-d unix|mac|pc ] [-h] [-r] FITS_file\n");
      exit(0);
    }

  i = 1;
  while (i < argc)
    { if (strcmp(argv[i],"-d")==0)
        { ++i;
          if (strcmp(argv[i],"unix") == 0)
            { delimiter = 0; }
          else if (strcmp(argv[i],"mac") == 0)
            { delimiter = 1; }
          else if (strcmp(argv[i],"pc") == 0)
            { delimiter = 2; }
          else
            { fprintf (stderr,"splitfits: Invalid delimiter type \"%s\".\n",
                               argv[i]);
              exit(10);
            }
        }

      else if (strcmp(argv[i],"-h")==0)
        { breakheader = 80; }

      else if (strcmp(argv[i],"-r")==0)
        { breakdata = 1; }

      else if (strcmp(argv[i],"-s")==0)
        { noisy = 0; }

      else
        { if ((ifp = fopen(argv[i],"rb"))==NULL)
            { fprintf (stderr,"splitfits: Unable to open \"%s\" for ",argv[i]);
              fprintf (stderr,"input (%s).\n",strerror(errno));
              exit(20);
            }
          strcpy(fitsname,argv[i]);
        }
      i++;
    }

  /* If we're adding line breaks, set up the buffering: */

  if (breakheader || breakdata)
    { write_block(NULL,NULL,-999,NULL,delimiter); }

  /*-----------------------------------------------------------------------*/

  /* Get the base filename to use for the output files: */

  sptr = strrchr(fitsname,'.');
  if (sptr != NULL)
    { sptr[0] = '\0'; }

  /* Read the first block of the input file.  We need to get some basic 
     information from this block to figure out how to proceed.
  */

  next_block(ifp, block, &blocknum);

  /* Call a routine to get the NAXIS1 value: */

  naxis1 = get_naxis1(block, blocknum);
  no_primary_data = (naxis1 == 0)? 1 : 0;
  naxis1 = (breakdata)? naxis1 : 0;


  /* Deal with primary header: */

  sprintf(filename,"%s.h01",fitsname);
  if ((ofp=fopen(filename,"wb")) == NULL)
    { fprintf (stderr,"splitfits: Unable to open \"%s\" for writing (%s).\n",
                      filename,strerror(errno));
      exit(30);
    }

  fileblock = 1;
  not_done = 1;
  while (not_done)
    { write_block(ofp, block, breakheader, filename, fileblock);

      if (end_block(block))
        { not_done = 0; }
      else
        { next_block(ifp, block, &blocknum); 
          fileblock++;
        }
    } 

  /* Close this file.  If there are no primary data, decrement the unitnum,
     so that the extension header will be appended to this header:
  */

  fclose(ofp);
  if (no_primary_data)
    { unitnum = 0; }
  else
    { fileblock = 0; 
      unitnum = 1;
    }   /* Reset local block counter for new file */


  /* Now we start a loop that should work for the rest of the file.  If the
     next block is an extension header, we start a new file and loop until
     we reach the END block.  If it is anything else, we loop through data
     blocks until the next extension block (or end of file).  There may be
     extension blocks without data;  these are written to separate files.
  */

  not_eof = next_block(ifp, block, &blocknum);

  while (not_eof)
    { fileblock++;


      if (strncmp(block,"XTENSION",8)==0)

      /* ...Extension header loop... */

      { /* First, generate the file name and open it: */

        unitnum++;
        sprintf (filename,"%s.h%2.2d",fitsname,unitnum);

        if (unitnum == 1  &&  no_primary_data)
          { strcpy(write_perm,"ab"); }  /* Append to Primary header only */
        else
          { strcpy(write_perm,"wb"); }

        if ((ofp=fopen(filename,write_perm)) == NULL)
          { fprintf (stderr,"splitfits: Unable to open \"%s\" for ",filename);
            fprintf (stderr,"writing (%s).\n",strerror(errno));
            exit(40);
          }

        /* Read the axis information if needed: */

        naxis1 = (breakdata)? get_naxis1(block, blocknum) : 0;


        /* Loop through header blocks: */

        not_done = 1;
        while (not_done)
          { write_block(ofp, block, breakheader, filename, fileblock); 
            if (end_block(block))
              { not_done = 0; }
            else
              { next_block(ifp, block, &blocknum);
                fileblock++;
              }
          }

        /* Close the file: */

        fclose(ofp);

        /* Next block: */

        not_eof  = next_block(ifp, block, &blocknum);
        fileblock = 0;
      }

    else

    /* ...Data section loop... */

      {  /* Generate the file name and open it: */

        sprintf (filename,"%s.d%2.2d",fitsname,unitnum);
        if ((ofp = fopen(filename,"wb")) == NULL)
        if ((ofp=fopen(filename,"ab")) == NULL)
          { fprintf (stderr,"splitfits: Unable to open \"%s\" for ",filename);
            fprintf (stderr,"writing (%s).\n",strerror(errno));
            exit(50);
          }

        /* Loop through data blocks: */

        not_done = 1;
        while (not_done)
          { write_block(ofp, block, naxis1, filename, fileblock);
            not_eof = next_block(ifp, block, &blocknum);
            if (not_eof)
               { if (strncmp(block,"XTENSION",8) == 0)
                   { not_done = 0; }
               }
            else
               { not_done = 0; }
            fileblock++;
          }

        /* If we're adding delimiters, make sure the last record has one: */

        write_block(ofp, NULL, -1, filename, fileblock);

        /* Close file (next block is already up): */

        fclose(ofp);
        fileblock = 0;
      }
    }

  exit(0);
}

/****************************************************************************/

int get_naxis1(char *block, int blocknum)

  /* Routine to read the given header block, assuming it is the first in 
     the header, and return the NAXIS1 keyword value, or 0 if there are
     no data associated with this header.  
  */

  { char  line[30];   /* Integer conversion buffer */
    int   value;      /* NAXIS1 value buffer */

    /* In either taype of block, the third line should be the NAXIS line,
       followed by NAXIS1 on the fourth, if there's any data.  Check for
       NAXIS1 first, since this is the more common case:

    */

    if (strncmp(block+240,"NAXIS1",6) == 0)
      { 
        /* Found it.  Get the number: */

        strncpy(line,block+250,20);
        line[20]='\0';
        value = atoi(line);
      }

    else if (strncmp(block+160,"NAXIS ",5) == 0)   /* Check for NAXIS = 0 */
      { strncpy(line,block+170,20);
        line[20]='\0';
        value = atoi(line);
        if (value != 0)
          { fprintf (stderr,"Invalid FITS file format in block %d.\n",
                             blocknum);
            exit(200);
          }
      }

    /* Done. */

    return value;

  }

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int end_block(char *block)

  /* Routine to search a header block to see if it contains an END statement.
     Returns true (1) if it does, false (0) otherwise.
  */

  { int  i;   /* loop variable */

    for (i=0; i<36; i++)
      { if (strncmp(block+(i*80),"END     ",8) == 0)
          { return 1; }
      }

    return 0;
  }

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int next_block(FILE *ifp, char *block, int *count)

  /* Routine to read in the next block from the FITS file and increment the
     input block counter.  Returns 1 if data were read successfully; 0
     otherwise.
  */

  { size_t  nbytes;   /* number of i/o bytes */

    (*count)++;

    nbytes = fread((void *) block, 1, 2880, ifp);
    if (nbytes == 0)
      { return 0; }

    else if (nbytes < 2880)
      { fprintf (stderr,"splitfits: Error attempting to read block %d ",
                        *count);
        fprintf (stderr,"from input file.\n");
        exit(300);
      }
    else
      { return 1; }
  }

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int write_block(FILE *ofp, char *block, int recsize,
                  char *filename, int fileblock)

  /* Routine to write the given block to the output file, adding carriage-
     control if indicated.  If a negative number is passed as the 'recsize'
     flag, the internal static buffer is flushed.
  */

  { static size_t bytes_left;     /* bytes left in current record, before CC */
    static char   delim[3];       /* record delimiter */
    static size_t delim_size;     /* size of delimiter */

    int   i,j,k;             /* Loop/subscript */
    char  *bp;               /* block pointer */


    /* Action depends on the recsize flag: */

    /* ...Initialize the static variables: */

    if (recsize == -999)
      { bytes_left = 0; 
        if (fileblock == 0)  /* 'fileblock' holds the delimiter flag */
          { delim[0] = '\n';
            delim_size = 1;
          }
        else if (fileblock == 1)
          { delim[0] = '\r';
            delim_size = 1;
          }
        else
          { delim[0] = '\r';
            delim[1] = '\n';
            delim_size = 2;
          }
        return 1;
      }

    /* ...Just write the block: */

    bp = block;

    if (recsize == 0)
      { if (fwrite((void *)block, 1, 2880, ofp) != 2880)
          { fprintf (stderr,"splitfits: Error writing block %d ",fileblock);
            fprintf (stderr,"to %s (%s).\n",filename,strerror(errno));
            exit(500);
          }
        return 1;
      }


    /* ...Write the block, adding carriage control: */

    else if (recsize > 0)

      { /* First, write any remaining bytes and delimiter on the current
           record:
        */

        if (bytes_left)  
          { fwrite((void *)block, 1, bytes_left, ofp);
            fwrite((void *)delim, 1, delim_size, ofp);
            bp += bytes_left;
            bytes_left = 2880 - bytes_left;
          }
        else
          { bytes_left = 2880; }

        /* Now, step through the block, adding delimiters as needed: */

        while (bytes_left >= recsize)
          { fwrite ((void *)bp, 1, recsize, ofp);
            fwrite ((void *)delim, 1, delim_size, ofp);
            bp += recsize;
            bytes_left -= recsize;
          }

        /* If there's a partial record left, write it and reset bytes_left
           to indicate the number of bytes from this record at the beginning
           of the next block:
        */

        if (bytes_left)
          { fwrite ((void *)bp, 1, bytes_left, ofp);
            bytes_left = recsize - bytes_left;
            bp = block;
          }

        /* Done. */

        return 1;
      }

    
    /* ...Write a final delimiter (no padding is added): */

    else
      { if (bytes_left)  /* Not needed otherwise */
          { fwrite ((void *)delim, 1, delim_size, ofp);
             bytes_left = 0;
             return 1;
          }
      }
  }


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
